# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

"""Convert SPN to JSON friendly dictionary."""

from fractions import Fraction

import scipy.stats
from sympy import E
from sympy import sqrt

from ..spn import ContinuousLeaf
from ..spn import DiscreteLeaf
from ..spn import NominalLeaf
from ..spn import ProductSPN
from ..spn import SumSPN

# Needed for "eval"
from ..sets import *
from ..transforms import Id
from ..transforms import Identity
from ..transforms import Radical
from ..transforms import Exponential
from ..transforms import Logarithm
from ..transforms import Abs
from ..transforms import Reciprocal
from ..transforms import Poly
from ..transforms import Piecewise
from ..transforms import EventInterval
from ..transforms import EventFiniteReal
from ..transforms import EventFiniteNominal
from ..transforms import EventOr
from ..transforms import EventAnd

def env_from_dict(env):
    if env is None:
        return None
    # Used in eval.
    return {eval(k): eval(v) for k, v in env.items()}

def env_to_dict(env):
    if len(env) == 1:
        return None
    return {repr(k): repr(v) for k, v in env.items()}

def scipy_dist_from_dict(dist):
    constructor = getattr(scipy.stats, dist['name'])
    return constructor(*dist['args'], **dist['kwds'])

def scipy_dist_to_dict(dist):
    return {
        'name': dist.dist.name,
        'args': dist.args,
        'kwds': dist.kwds
    }

def spn_from_dict(metadata):
    if metadata['class'] == 'NominalLeaf':
        symbol = Id(metadata['symbol'])
        dist = {x: Fraction(w[0], w[1]) for x, w in metadata['dist']}
        return NominalLeaf(symbol, dist)
    if metadata['class'] == 'ContinuousLeaf':
        symbol = Id(metadata['symbol'])
        dist = scipy_dist_from_dict(metadata['dist'])
        support = eval(metadata['support'])
        conditioned = metadata['conditioned']
        env = env_from_dict(metadata['env'])
        return ContinuousLeaf(symbol, dist, support, conditioned, env=env)
    if metadata['class'] == 'DiscreteLeaf':
        symbol = Id(metadata['symbol'])
        dist = scipy_dist_from_dict(metadata['dist'])
        support = eval(metadata['support'])
        conditioned = metadata['conditioned']
        env = env_from_dict(metadata['env'])
        return DiscreteLeaf(symbol, dist, support, conditioned, env=env)
    if metadata['class'] == 'SumSPN':
        children = [spn_from_dict(c) for c in metadata['children']]
        weights = metadata['weights']
        return SumSPN(children, weights)
    if metadata['class'] == 'ProductSPN':
        children = [spn_from_dict(c) for c in metadata['children']]
        return ProductSPN(children)

    assert False, 'Cannot convert %s to SPN' % (metadata,)

def spn_to_dict(spn):
    if isinstance(spn, NominalLeaf):
        return {
            'class'        : 'NominalLeaf',
            'symbol'       : spn.symbol.token,
            'dist'         : [
                (str(x), (w.numerator, w.denominator))
                for x, w in spn.dist.items()
            ],
            'env'         : env_to_dict(spn.env),
        }
    if isinstance(spn, ContinuousLeaf):
        return {
            'class'         : 'ContinuousLeaf',
            'symbol'        : spn.symbol.token,
            'dist'          : scipy_dist_to_dict(spn.dist),
            'support'       : repr(spn.support),
            'conditioned'   : spn.conditioned,
            'env'           : env_to_dict(spn.env),
        }
    if isinstance(spn, DiscreteLeaf):
        return {
            'class'         : 'DiscreteLeaf',
            'symbol'        : spn.symbol.token,
            'dist'          : scipy_dist_to_dict(spn.dist),
            'support'       : repr(spn.support),
            'conditioned'   : spn.conditioned,
            'env'           : env_to_dict(spn.env),
        }
    if isinstance(spn, SumSPN):
        return {
            'class'         : 'SumSPN',
            'children'      : [spn_to_dict(c) for c in spn.children],
            'weights'       : spn.weights,
        }
    if isinstance(spn, ProductSPN):
        return {
            'class'         : 'ProductSPN',
            'children'      : [spn_to_dict(c) for c in spn.children],
        }
    assert False, 'Cannot convert %s to JSON' % (spn,)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import log

from sppl.compilers.ast_to_spn import Id
from sppl.compilers.ast_to_spn import IfElse
from sppl.compilers.ast_to_spn import Otherwise
from sppl.compilers.ast_to_spn import Sample
from sppl.compilers.ast_to_spn import Sequence
from sppl.compilers.ast_to_spn import Transform
from sppl.distributions import bernoulli
from sppl.distributions import norm
from sppl.math_util import allclose

X = Id('X')
Y = Id('Y')
Z = Id('Z')

def test_simple_transform():
    command = Sequence(
        Sample(X, norm(loc=0, scale=1)),
        Transform(Z, X**2))
    model = command.interpret()
    assert model.get_symbols() == {Z, X}
    assert model.env == {Z:X**2, X:X}
    assert (model.logprob(Z > 0)) == 0

def test_if_else_transform():
    model = Sequence(
        Sample(X, norm(loc=0, scale=1)),
        IfElse(
            X > 0,      Transform(Z, X**2),
            Otherwise,  Transform(Z, X))).interpret()
    assert model.children[0].env == {X:X, Z:X**2}
    assert model.children[1].env == {X:X, Z:X}
    assert allclose(model.children[0].logprob(Z > 0), 0)
    assert allclose(model.children[1].logprob(Z > 0), -float('inf'))
    assert allclose(model.logprob(Z > 0), -log(2))

def test_if_else_transform_reverse():
    command = Sequence(
        Sample(X, norm(loc=0, scale=1)),
        Sample(Y, bernoulli(p=0.5)),
        IfElse(
            Y << {0},  Transform(Z, X**2),
            Otherwise, Transform(Z, X)))
    model = command.interpret()
    assert allclose(model.logprob(Z > 0), log(3) - log(4))



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import exp

from .spn import RealLeaf
from .spn import DiscreteLeaf
from .spn import LeafSPN
from .spn import NominalLeaf
from .spn import ProductSPN
from .spn import SumSPN

def render_nested_lists_concise(spn):
    if isinstance(spn, LeafSPN):
        return [(str(k), str(v)) for k, v in spn.env.items()]
    if isinstance(spn, SumSPN):
        return ['+(%d)' % (len(spn.children),),
            # [exp(w) for w in spn.weights],
            [render_nested_lists_concise(c) for c in spn.children]
        ]
    if isinstance(spn, ProductSPN):
        return ['*(%d)' % (len(spn.children),),
            [render_nested_lists_concise(c) for c in spn.children]
        ]

def render_nested_lists(spn):
    if isinstance(spn, NominalLeaf):
        return ['NominalLeaf', [
            ['symbol', spn.symbol],
            ['env', dict(spn.env)],
            ['dist', {str(x): float(w) for x, w in spn.dist.items()}]]
        ]
    if isinstance(spn, RealLeaf):
        return ['RealLeaf', [
            ['symbol', spn.symbol],
            ['env', dict(spn.env)],
            ['dist', (spn.dist.dist.name, spn.dist.args, spn.dist.kwds)],
            ['support', spn.support],
            ['conditioned', spn.conditioned]]
        ]
    if isinstance(spn, DiscreteLeaf):
        return ['DiscreteLeaf', [
            ['symbol', spn.symbol],
            ['dist', (spn.dist.dist.name, spn.dist.args, spn.dist.kwds)],
            ['support', spn.support],
            ['conditioned', spn.conditioned]]
        ]
    if isinstance(spn, SumSPN):
        return ['SumSPN', [
            ['symbols', list(spn.symbols)],
            ['weights', [exp(w) for w in spn.weights]],
            ['n_children', len(spn.children)],
            ['children', [render_nested_lists(c) for c in spn.children]]]
        ]
    if isinstance(spn, ProductSPN):
        return ['ProductSPN', [
            ['symbols', list(spn.symbols)],
            ['n_children', len(spn.children)],
            ['children', [render_nested_lists(c) for c in spn.children]]]
        ]
    assert False, 'Unknown SPN type: %s' % (spn,)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest

from sppl.distributions import choice
from sppl.distributions import norm
from sppl.distributions import poisson
from sppl.math_util import allclose
from sppl.transforms import Id

def test_transform_real_leaf_logprob():
    X = Id('X')
    Y = Id('Y')
    Z = Id('Z')
    spn = (X >> norm(loc=0, scale=1))

    with pytest.raises(AssertionError):
        spn.transform(Z, Y**2)
    with pytest.raises(AssertionError):
        spn.transform(X, X**2)

    spn = spn.transform(Z, X**2)
    assert spn.env == {X:X, Z:X**2}
    assert spn.get_symbols() == {X, Z}
    assert spn.logprob(Z < 1) == spn.logprob(X**2 < 1)
    assert spn.logprob((Z < 1) | ((X + 1) < 3)) \
        == spn.logprob((X**2 < 1) | ((X+1) < 3))

    spn = spn.transform(Y, 2*Z)
    assert spn.env == {X:X, Z:X**2, Y:2*Z}
    assert spn.logprob(Y**(1,3) < 10) \
        == spn.logprob((2*Z)**(1,3) < 10) \
        == spn.logprob((2*(X**2))**(1,3) < 10) \

    W = Id('W')
    spn = spn.transform(W, X > 1)
    assert allclose(spn.logprob(W), spn.logprob(X > 1))

def test_transform_real_leaf_sample():
    X = Id('X')
    Z = Id('Z')
    Y = Id('Y')
    spn = (X >> poisson(loc=-1, mu=1))
    spn = spn.transform(Z, X+1)
    spn = spn.transform(Y, Z-1)
    samples = spn.sample(100)
    assert any(s[X] == -1 for s in samples)
    assert all(0 <= s[Z] for s in samples)
    assert all(s[Y] == s[X] for s in samples)
    assert all(spn.sample_func(lambda X,Y,Z: X-Y+Z==Z, 100))
    assert all(set(s) == {X,Y} for s in spn.sample_subset([X, Y], 100))

def test_transform_sum():
    X = Id('X')
    Z = Id('Z')
    Y = Id('Y')
    spn \
        = 0.3*(X >> norm(loc=0, scale=1)) \
        | 0.7*(X >> choice({'0': 0.4, '1': 0.6}))
    with pytest.raises(Exception):
        # Cannot transform Nominal variate.
        spn.transform(Z, X**2)
    spn \
        = 0.3*(X >> norm(loc=0, scale=1)) \
        | 0.7*(X >> poisson(mu=2))
    spn = spn.transform(Z, X**2)
    assert spn.logprob(Z < 1) == spn.logprob(X**2 < 1)
    assert spn.children[0].env == spn.children[1].env
    spn = spn.transform(Y, Z/2)
    assert spn.children[0].env \
        == spn.children[1].env \
        == {X:X, Z:X**2, Y:Z/2}

def test_transform_product():
    X = Id('X')
    Y = Id('Y')
    W = Id('W')
    Z = Id('Z')
    V = Id('V')
    spn \
        = (X >> norm(loc=0, scale=1)) \
        & (Y >> poisson(mu=10))
    with pytest.raises(Exception):
        # Cannot use symbols from different transforms.
        spn.transform(W, (X > 0) | (Y << {'0'}))
    spn = spn.transform(W, (X**2 - 3*X)**(1,10))
    spn = spn.transform(Z, (W > 0) | (X**3 < 1))
    spn = spn.transform(V, Y/10)
    assert allclose(
        spn.logprob(W>1),
        spn.logprob((X**2 - 3*X)**(1,10) > 1))
    with pytest.raises(Exception):
        spn.tarnsform(Id('R'), (V>1) | (W < 0))



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import isinf

import numpy

from scipy.special import logsumexp

# Implementation of log1mexp and logdiffexp from PyMC3 math module.
# https://github.com/pymc-devs/pymc3/blob/master/pymc3/math.py
def log1mexp(x):
    if x < 0.683:
        return numpy.log(-numpy.expm1(-x))
    else:
        return numpy.log1p(-numpy.exp(-x))

def logdiffexp(a, b):
    if b < a:
        return a + log1mexp(a - b)
    if allclose(b, a):
        return -float('inf')
    raise ValueError('Negative term in logdiffexp.')

def lognorm(array):
    M = logsumexp(array)
    return [a - M for a in array]

def logflip(logp, array, size, rng):
    p = numpy.exp(lognorm(logp))
    return flip(p, array, size, rng)

def flip(p, array, size, rng):
    p = normalize(p)
    return random(rng).choice(array, size=size, p=p)

def normalize(p):
    s = float(sum(p))
    return numpy.asarray(p, dtype=float) / s

def allclose(values, x):
    return numpy.allclose(values, x)

def isinf_pos(x):
    return isinf(x) and x > 0

def isinf_neg(x):
    return isinf(x) and x < 0

def random(x):
    return x or numpy.random

int_or_isinf_neg = lambda a: isinf_neg(a) or float(a) == int(a)
int_or_isinf_pos = lambda a: isinf_pos(a) or float(a) == int(a)
float_to_int = lambda a: a if isinf(a) else int(a)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest

from sppl.transforms import Id
from sppl.transforms import Logarithm

X = Id('X')
Z = Id('Z')
Y = Id('Y')

b = (1, 10)
cases = [
    # Basic cases.
    [X                                   , {}       , X],
    [1/X                                 , {}       , 1/X],
    [X**2+X                              , {}       , X**2+X],
    [X                                   , {}       , X],
    [X                                   , {Z:X}    , X],
    [Z                                   , {Z:2**X} , 2**X],
    [((Z+1)**b)                          , {Z:1/X}  , (1/X+1)**b],
    [((2**Z+1)**b)                       , {Z:X+1}  , (2**(X+1)+1)**b],
    [((2**Logarithm(Z, 2)+1)**b)         , {Z:X+1}  , (2**Logarithm((X+1), 2)+1)**b] ,
    [((2**abs(Logarithm(Z, 2))+1)**b)    , {Z:X+1}  , (2**abs(Logarithm((X+1), 2))+1)**b],
    [((2**abs(Logarithm(1/Z, 2))+1)**b)  , {Z:X+1}  , (2**abs(Logarithm(1/(X+1), 2))+1)**b],
    # Compound cases.
    [(Z > 1)*(1/Z) + (Z < 1)*Z**b
        , {Z:X+1}
        , ((X+1) > 1)*(1/(X+1)) + ((X+1) < 1)*(X+1)**b],
    [(Z << {'a'}) & (Y < 3)
        , {Y:1/X}
        , (Z << {'a'}) & (1/X < 3)],
    [((Y > 1) | Z << {'a'}) & (Y < 3)
        , {Z : Y**2, Y:1/X}
        , (((1/X) > 1) | ((1/X)**2) << {'a'}) & ((1/X) < 3)],
]
@pytest.mark.parametrize('case', cases)
def test_substitute_basic(case):
    (expr, env, expr_prime) = case
    assert expr.substitute(env) == expr_prime

@pytest.mark.parametrize('case', cases)
def test_substitute_transitive(case):
    (expr, env, expr_prime) = case
    if len(env) == 1:
        [(s0, s1)] = env.items()
        s2 = Id('s2')
        env_prime = {s0: s2, s2: s1}
        assert expr.substitute(env_prime) == expr_prime



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest

from sympy import exp as SymExp
from sympy import log as SymLog
from sympy import symbols

from sppl.sets import FiniteReal
from sppl.sym_util import get_symbols
from sppl.sym_util import partition_finite_real_contiguous
from sppl.sym_util import partition_list_blocks

(X0, X1, X2, X3, X4, X5, X6, X7, X8, X9) = symbols('X:10')

def test_get_symbols():
    syms = get_symbols((X0 > 3) & (X1 < 4))
    assert len(syms) == 2
    assert X0 in syms
    assert X1 in syms

    syms = get_symbols((SymExp(X0) > SymLog(X1)+10) & (X2 < 4))
    assert len(syms) == 3
    assert X0 in syms
    assert X1 in syms
    assert X2 in syms

@pytest.mark.parametrize('a, b', [
    ([0, 1, 2, 3], [[0], [1], [2], [3]]),
    ([0, 1, 2, 1], [[0], [1, 3], [2]]),
    (['0', '0', 2, '0'], [[0, 1, 3], [2]]),
])
def test_partition_list_blocks(a, b):
    solution = partition_list_blocks(a)
    assert solution == b

@pytest.mark.parametrize('a, b', [
    (FiniteReal(0,1,2), [FiniteReal(0,1,2)]),
    (FiniteReal(0,3,1,2), [FiniteReal(0,1,2,3)]),
    (FiniteReal(-1,3,1,2), [FiniteReal(-1), FiniteReal(1,2,3)]),
    (FiniteReal(-1,3,1,2,-2,-7), [FiniteReal(-7), FiniteReal(-1,-2), FiniteReal(1,2,3)]),
    (FiniteReal(-1,3,1,2,-2,-7,0), [FiniteReal(-7), FiniteReal(-2,-1,0,1,2,3)]),
])
def test_parition_finite_real_contiguous(a, b):
    solution = partition_finite_real_contiguous(a)
    assert solution == b



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from fractions import Fraction
from math import log

import numpy
import pytest

from sppl.distributions import choice
from sppl.math_util import allclose
from sppl.math_util import isinf_neg
from sppl.sets import FiniteNominal
from sppl.transforms import Id

def test_nominal_distribution():
    X = Id('X')
    spn = X >> choice({
        'a': Fraction(1, 5),
        'b': Fraction(1, 5),
        'c': Fraction(3, 5),
    })
    assert allclose(spn.logprob(X << {'a'}), log(Fraction(1, 5)))
    assert allclose(spn.logprob(X << {'b'}), log(Fraction(1, 5)))
    assert allclose(spn.logprob(X << {'a', 'c'}), log(Fraction(4, 5)))
    assert allclose(
        spn.logprob((X << {'a'}) & ~(X << {'b'})),
        log(Fraction(1, 5)))
    assert allclose(
        spn.logprob((X << {'a', 'b'}) & ~(X << {'b'})),
        log(Fraction(1, 5)))
    assert spn.logprob((X << {'d'})) == -float('inf')
    assert spn.logprob((X << ())) == -float('inf')

    samples = spn.sample(100)
    assert all(s[X] in spn.support for s in samples)

    samples = spn.sample_subset([X], 100)
    assert all(len(s)==1 and s[X] in spn.support for s in samples)

    with pytest.raises(Exception):
        spn.sample_subset(['f'], 100)

    predicate = lambda X: (X in {'a', 'b'}) or X in {'c'}
    samples = spn.sample_func(predicate, 100)
    assert all(samples)

    predicate = lambda X: (not (X in {'a', 'b'})) and (not (X in {'c'}))
    samples = spn.sample_func(predicate, 100)
    assert not any(samples)

    func = lambda X: 1 if X in {'a'} else None
    samples = spn.sample_func(func, 100, prng=numpy.random.RandomState(1))
    assert sum(1 for s in samples if s == 1) > 12
    assert sum(1 for s in samples if s is None) > 70

    with pytest.raises(ValueError):
        spn.sample_func(lambda Y: Y, 100)

    spn_condition = spn.condition(X<<{'a', 'b'})
    assert spn_condition.support == FiniteNominal('a', 'b', 'c')
    assert allclose(spn_condition.logprob(X << {'a'}), -log(2))
    assert allclose(spn_condition.logprob(X << {'b'}), -log(2))
    assert spn_condition.logprob(X << {'c'}) == -float('inf')

    assert isinf_neg(spn_condition.logprob(X**2 << {1}))

    with pytest.raises(ValueError):
        spn.condition(X << {'python'})
    assert spn.condition(~(X << {'python'})) == spn



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from functools import reduce
from itertools import chain
from itertools import combinations

from .sets import EmptySet
from .transforms import EventAnd
from .transforms import EventBasic
from .transforms import EventOr
from .transforms import Id

def dnf_factor(event, lookup=None):
    # Given an event (in DNF) and a dictionary lookup mapping symbols
    # to integers, this function returns a list R of dictionaries
    # R[i][j] is a conjunction events in the i-th DNF clause whose symbols
    # are assigned to integer j in the lookup dictionary.
    #
    # For example, if e is any predicate
    # event = (e(X0) & e(X1) & ~e(X2)) | (~e(X1) & e(X2) & e(X3) & e(X4)))
    # lookup = {X0: 0, X1: 1, X2: 0, X3: 1, X4: 2}
    # The output is
    # R = [
    #   { // First clause
    #       0: e(X0) & ~e(X2),
    #       1: e(X1)},
    #   { // Second clause
    #       0: e(X2),
    #       1: ~e(X1) & e(X3)},
    #       2: e(X4)},
    # ]
    if lookup is None:
        lookup = {s:s for s in event.get_symbols()}

    if isinstance(event, EventBasic):
        # Literal.
        symbols = event.get_symbols()
        assert len(symbols) == 1
        symbol = list(symbols)[0]
        key = lookup[symbol]
        return ({key: event},)

    if isinstance(event, EventAnd):
        # Conjunction.
        assert all(isinstance(e, EventBasic) for e in event.subexprs)
        mappings = (dnf_factor(e, lookup) for e in event.subexprs)
        events = {}
        for mapping in mappings:
            for m in mapping:
                for key, ev in m.items():
                    if key not in events:
                        events[key] = ev
                    else:
                        events[key] &= ev
        return (events,)

    if isinstance(event, EventOr):
        # Disjunction.
        assert all(isinstance(e, (EventAnd, EventBasic)) for e in event.subexprs)
        mappings = (dnf_factor(e, lookup) for e in event.subexprs)
        return tuple(chain.from_iterable(mappings))

    assert False, 'Invalid DNF event: %s' % (event,)

def dnf_normalize(event):
    if isinstance(event, EventBasic):
        if isinstance(event.subexpr, Id):
            return event
    # Given an arbitrary event, rewrite in terms of only Id by
    # solving the subexpressions and return the resulting DNF formula,
    # or None if all solutions evaluate to EmptySet.
    event_dnf = event.to_dnf()
    event_factor = dnf_factor(event_dnf)
    solutions = list(filter(lambda x: all(y[1] is not EmptySet for y in x), [
        [(symbol, ev.solve()) for symbol, ev in clause.items()]
        for clause in event_factor
    ]))
    if not solutions:
        return None
    conjunctions = [
        reduce(lambda x, e: x & e, [(symbol << S) for symbol, S in clause])
        for i, clause in enumerate(solutions) if clause not in solutions[:i]
    ]
    disjunctions = reduce(lambda x, e: x|e, conjunctions)
    return disjunctions.to_dnf()

def dnf_non_disjoint_clauses(event):
    # Given an event in DNF, returns a dictionary R
    # such that R[j] = [i | i < j and event[i] intersects event[j]]
    event_factor = dnf_factor(event)
    solutions = [
        {symbol: ev.solve() for symbol, ev in clause.items()}
        for clause in event_factor
    ]

    n_clauses = len(event_factor)
    overlap_dict = {}
    for i, j in combinations(range(n_clauses), 2):
        # Exit if any symbol in i does not intersect a symbol in j.
        intersections = (
            solutions[i][symbol] & solutions[j][symbol]
                if (symbol in solutions[j]) else
                solutions[i][symbol]
            for symbol in solutions[i]
        )
        if any(x is EmptySet for x in intersections):
            continue
        # Exit if any symbol in j is EmptySet.
        if any(solutions[j] is EmptySet for symbol in solutions[j]):
            continue
        # All symbols intersect, so clauses overlap.
        if j not in overlap_dict:
            overlap_dict[j] = []
        overlap_dict[j].append(i)

    return overlap_dict

def dnf_to_disjoint_union(event):
    # Given an event in DNF, returns an event in DNF where all the
    # clauses are disjoint from one another, by recursively solving the
    # identity E = (A or B or C) = (A) or (B and ~A) or (C and ~A and ~B).
    # Base case.
    if isinstance(event, (EventBasic, EventAnd)):
        return event
    # Find indexes of pairs of clauses that overlap.
    overlap_dict = dnf_non_disjoint_clauses(event)
    if not overlap_dict:
        return event
    # Create the cascading negated clauses.
    n_clauses = len(event.subexprs)
    clauses_disjoint = [
        reduce(
            lambda state, event: state & ~event,
            (event.subexprs[j] for j in overlap_dict.get(i, [])),
            event.subexprs[i])
        for i in range(n_clauses)
    ]
    # Recursively find the solutions for each clause.
    clauses_normalized = [dnf_normalize(clause) for clause in clauses_disjoint]
    solutions = [dnf_to_disjoint_union(c) for c in clauses_normalized if c]
    # Return the merged solution.
    return reduce(lambda a, b: a|b, solutions)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

"""Convert SPN to SPPL."""

from io import StringIO
from math import exp

from ..spn import RealLeaf
from ..spn import NominalLeaf
from ..spn import ProductSPN
from ..spn import SumSPN

get_indentation = lambda i: ' ' * i
float_to_str = lambda x,fw: '%1.*f' % (fw, float(x)) if fw else str(x)
class _SPPL_Render_State:
    def __init__(self, stream=None, branches=None, indentation=None,
            fwidth=None):
        self.stream = stream or StringIO()
        self.branches = branches or []
        self.indentation = indentation or 0
        self.fwidth = fwidth
def render_sppl_choice(symbol, dist, stream, indentation, fwidth):
    idt = get_indentation(indentation)
    dist_pairs = [
        '\'%s\' : %s' % (k, float_to_str(v, fwidth))
        for (k, v) in dist.items()
    ]
    # Write each outcome on new line.
    if sum(len(x) for x in dist_pairs) > 30:
        idt4 = get_indentation(indentation + 4)
        prefix = ',\n%s' % (idt4,)
        dist_op = '\n%s' % (idt4,)
        dist_cl = '\n%s' % (idt,)
    # Write all outcomes on same line.
    else:
        prefix = ', '
        dist_op = ''
        dist_cl = ''
    dist_str = prefix.join(dist_pairs)
    stream.write('%s%s ~= choice({%s%s%s})' %
        (idt, symbol, dist_op, dist_str, dist_cl))
    stream.write('\n')
def render_sppl_helper(spn, state):
    if isinstance(spn, NominalLeaf):
        assert len(spn.env) == 1
        render_sppl_choice(
            spn.symbol,
            spn.dist,
            state.stream,
            state.indentation,
            state.fwidth)
        return state
    if isinstance(spn, RealLeaf):
        kwds = ', '.join([
            '%s=%s' % (k, float_to_str(v, state.fwidth))
            for k, v in spn.dist.kwds.items()
        ])
        dist = '%s(%s)' % (spn.dist.dist.name, kwds)
        idt = get_indentation(state.indentation)
        state.stream.write('%s%s ~= %s' % (idt, spn.symbol, dist))
        state.stream.write('\n')
        if spn.conditioned:
            event = spn.symbol << spn.support
            # TODO: Consider using repr(event)
            state.stream.write('%scondition(%s)' % (idt, event))
            state.stream.write('\n')
        for i, (var, expr) in enumerate(spn.env.items()):
            if 1 <= i:
                state.stream.write('%s%s ~= %s' % (idt, var, expr))
                state.stream.write('\n')
        return state
    if isinstance(spn, ProductSPN):
        for child in spn.children:
            state = render_sppl_helper(child, state)
        return state
    if isinstance(spn, SumSPN):
        if len(spn.children) == 0:
            return state
        if len(spn.children) == 1:
            return render_sppl_helper(spn.children[0], state)
        branch_var = 'branch_var_%s' % (len(state.branches))
        branch_idxs = [str(i) for i in range(len(spn.children))]
        branch_dist = {k: exp(w) for k, w in zip(branch_idxs, spn.weights)}
        state.branches.append((branch_var, branch_dist))
        # Write the branches.
        for i, child in zip(branch_idxs, spn.children):
            ifstmt = 'if' if i == '0' else 'elif'
            idt = get_indentation(state.indentation)
            state.stream.write('%s%s (%s == \'%s\'):'
                % (idt, ifstmt, branch_var, i))
            state.stream.write('\n')
            state.indentation += 4
            state = render_sppl_helper(child, state)
            state.stream.write('\n')
            state.indentation -= 4
        return state
    assert False, 'Unknown spn %s' % (spn,)

def render_sppl(spn, stream=None, fwidth=None):
    if stream is None:
        stream = StringIO()
    state = _SPPL_Render_State(fwidth=fwidth)
    state = render_sppl_helper(spn, state)
    assert state.indentation == 0
    # Write the import.
    stream.write('from sppl.distributions import *')
    stream.write('\n')
    stream.write('\n')
    # Write the branch variables (if any).
    for branch_var, branch_dist in state.branches:
        render_sppl_choice(branch_var, branch_dist, stream, 0, fwidth)
    stream.write('\n')
    # Write the SPPL.
    stream.write(state.stream.getvalue())
    return stream



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from .magics import SPPL_Magics

def load_ipython_extension(ipython):
    magics = SPPL_Magics(ipython)
    ipython.register_magics(magics)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import os

from itertools import chain
from math import isinf

import sympy

from sympy.calculus.util import limit

from .sets import EmptySet
from .sets import ExtReals
from .sets import FiniteReal
from .sets import Interval
from .sets import Reals
from .sets import convert_sympy
from .sets import oo
from .sets import make_union
from .sym_util import get_symbols
from .timeout import timeout

TIMEOUT_SYMBOLIC = 5

def get_poly_symbol(expr):
    symbols = tuple(get_symbols(expr))
    assert len(symbols) == 1
    return symbols[0]

# ==============================================================================
# Solving inequalities.

def solve_poly_inequality(expr, b, strict, extended=None):
    # Handle infinite case.
    if isinf(b):
        return solve_poly_inequality_inf(expr, b, strict, extended=extended)
    # Bypass symbolic inference.
    if os.environ.get('SPN_NO_SYMBOLIC'):
        return solve_poly_inequality_numerically(expr, b, strict)
    # Solve symbolically, if possible.
    try:
        with timeout(seconds=TIMEOUT_SYMBOLIC):
            result_symbolic = solve_poly_inequality_symbolically(expr, b, strict)
    except TimeoutError:
        result_symbolic = None
    if result_symbolic is not None:
        if not isinstance(result_symbolic, (sympy.ConditionSet, sympy.Intersection)):
            return convert_sympy(result_symbolic)
    # Solve numerically.
    return solve_poly_inequality_numerically(expr, b, strict)

def solve_poly_inequality_symbolically(expr, b, strict):
    expr = (expr < b) if strict else (expr <= b)
    return sympy.solveset(expr, domain=sympy.Reals)

def solve_poly_inequality_numerically(expr, b, strict):
    poly = expr - b
    symX = get_poly_symbol(expr)
    # Obtain numerical roots.
    roots = sympy.nroots(poly)
    zeros = sorted([r for r in roots if r.is_real])
    if not zeros:
        return sympy.EmptySet
    # Construct intervals around roots.
    mk_intvl = lambda a, b: \
        Interval(a, b, left_open=strict, right_open=strict)
    intervals = list(chain(
        [mk_intvl(-oo, zeros[0])],
        [mk_intvl(x, y) for x, y in zip(zeros, zeros[1:])],
        [mk_intvl(zeros[-1], oo)]))
    # Define probe points.
    xs_probe = list(chain(
        [zeros[0] - 1/2],
        [(i.left + i.right)/2 for i in intervals[1:-1]
            if isinstance(i, Interval)],
        [zeros[-1] + 1/2]))
    # Evaluate poly at the probe points.
    f_xs_probe = [poly.subs(symX, x) for x in xs_probe]
    # Return intervals where poly is less than zero.
    idxs = [i for i, fx in enumerate(f_xs_probe) if fx < 0]
    return make_union(*[intervals[i] for i in idxs])

def solve_poly_inequality_inf(expr, b, strict, extended=None):
    # Minimum value of polynomial is negative infinity.
    assert isinf(b)
    ext = True if extended is None else extended
    if b < 0:
        if strict or not ext:
            return EmptySet
        else:
            return solve_poly_equality_inf(expr, b)
    # Maximum value of polynomial is positive infinity.
    else:
        if strict:
            xinf = solve_poly_equality_inf(expr, -oo) if ext else EmptySet
            return Reals | xinf
        else:
            return ExtReals if ext else Reals

# ==============================================================================
# Solving equalities.

def solve_poly_equality(expr, b):
    # Handle infinite case.
    if isinf(b):
        return solve_poly_equality_inf(expr, b)
    # Bypass symbolic inference.
    if os.environ.get('SPN_NO_SYMBOLIC'):
        return solve_poly_equality_numerically(expr, b)
    # Solve symbolically, if possible.
    try:
        with timeout(seconds=TIMEOUT_SYMBOLIC):
            result_symbolic = solve_poly_equality_symbolically(expr, b)
    except TimeoutError:
        result_symbolic = None
    if result_symbolic is not None:
        if not isinstance(result_symbolic, (sympy.ConditionSet, sympy.Intersection)):
            return convert_sympy(result_symbolic)
    # Solve numerically.
    return solve_poly_equality_numerically(expr, b)

def solve_poly_equality_symbolically(expr, b):
    soln_lt = solve_poly_inequality_symbolically(expr, b, False)
    soln_gt = solve_poly_inequality_symbolically(-expr, -b, False)
    return sympy.Intersection(soln_lt, soln_gt)

def solve_poly_equality_numerically(expr, b):
    roots = sympy.nroots(expr-b)
    zeros = [r for r in roots if r.is_real]
    return FiniteReal(*zeros)

def solve_poly_equality_inf(expr, b):
    assert isinf(b)
    symX = get_poly_symbol(expr)
    val_pos_inf = limit(expr, symX, oo)
    val_neg_inf = limit(expr, symX, -oo)
    check_equal = lambda x: isinf(x) and ((x > 0) if (b > 0) else (x < 0))
    if check_equal(val_pos_inf) and check_equal(val_neg_inf):
        return FiniteReal(oo, -oo)
    if check_equal(val_pos_inf):
        return FiniteReal(oo)
    if check_equal(val_neg_inf):
        return FiniteReal(-oo)
    return EmptySet



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import signal

class timeout:
    # Adapted from https://stackoverflow.com/a/22348885
    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message
    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)
    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)
    def __exit__(self, type, value, traceback):
        signal.alarm(0)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import sys
from collections import namedtuple

from IPython.core.magic import Magics
from IPython.core.magic import cell_magic
from IPython.core.magic import line_magic
from IPython.core.magic import magics_class
from IPython.core.magic import needs_local_scope

from sppl.compilers.sppl_to_python import SPPL_Compiler

from .render import render_graphviz

Model = namedtuple('Model', ['source', 'compiler', 'namespace'])

@magics_class
class SPPL_Magics(Magics):

    def __init__(self, shell):
        super().__init__(shell)
        self.programs = {}

    @line_magic
    def sppl_get_spn(self, line):
        assert line in self.programs, 'unknown program %s' % (line,)
        return getattr(self.programs[line].namespace, line)

    @cell_magic
    def sppl(self, line, cell):
        if not line:
            sys.stderr.write('specify model name after %%sppl')
            return
        if line in self.programs:
            del self.programs[line]
        compiler = SPPL_Compiler(cell, line)
        namespace = compiler.execute_module()
        self.programs[line] = Model(cell, compiler, namespace)

    @line_magic
    def sppl_to_python(self, line):
        assert line in self.programs, 'unknown program %s' % (line,)
        print(self.programs[line].compiler.render_module())

    @needs_local_scope
    @line_magic
    def sppl_to_graph(self, line, local_ns):
        tokens = line.strip().split(' ')
        line = tokens[0]
        filename = tokens[1] if len(tokens) == 2 else None
        if line in self.programs:
            spn = self.sppl_get_spn(line)
        elif line in local_ns:
            spn = local_ns[line]
        else:
            assert False, 'unknown program %s' % (line,)
        return render_graphviz(spn, filename=filename)

    @line_magic
    def sppl_get_namespace(self, line):
        assert line in self.programs, 'unknown program %s' % (line,)
        return self.programs[line].namespace



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest

from sppl.dnf import dnf_factor
from sppl.dnf import dnf_non_disjoint_clauses
from sppl.dnf import dnf_to_disjoint_union

from sppl.transforms import EventOr
from sppl.transforms import Exp
from sppl.transforms import Id
from sppl.transforms import Log
from sppl.transforms import Sqrt

(X0, X1, X2, X3, X4, X5) = [Id("X%d" % (i,)) for i in range(6)]

events = [
    X0 < 0,
    (X1<<(0,1)) & (X2<0),
    (X1<0) | (X2<0),
    (X0<0) | ((X1<0) & ~(X2<0)),
    ((X0<0) & ~(0<X2)) | ((X1<0) & ~(X2<0)),
]
@pytest.mark.parametrize('event', events)
def test_to_dnf_no_change(event):
    assert event.to_dnf() == event

def test_to_dnf_changes():
    A = X0 < 0
    B = X1 < 0
    C = X2 < 0
    D = X3 < 100
    E = X4 < 10

    event = A & (B | C)
    event_dnf = event.to_dnf()
    assert len(event_dnf.subexprs) == 2
    assert A & B in event_dnf.subexprs
    assert A & C in event_dnf.subexprs

    event = (A & (B | C)) | D
    event_dnf = event.to_dnf()
    assert len(event_dnf.subexprs) == 3
    assert A & B in event_dnf.subexprs
    assert A & C in event_dnf.subexprs
    assert D in event_dnf.subexprs

    event = (A & (B | C)) | (D & E)
    event_dnf = event.to_dnf()
    assert len(event_dnf.subexprs) == 3
    assert A & B in event_dnf.subexprs
    assert A & C in event_dnf.subexprs
    assert D & E in event_dnf.subexprs

    event = (A | (B | C)) & (D & E)
    event_dnf = event.to_dnf()
    assert A & D & E in event_dnf.subexprs
    assert B & D & E in event_dnf.subexprs
    assert C & D & E in event_dnf.subexprs

def test_to_dnf_invert():
    A = X0 < 0
    B = X1 < 0
    C = X2 < 0
    D = X3 < 0

    event = ~(A | B | C)
    assert event.to_dnf() == ~A & ~B & ~C

    event = ~(A | B | ~C)
    assert event.to_dnf() == ~A & ~B & C

    event = ~((A | B | C) & D)
    #  =>  ~(A | B | C) | ~D
    assert event.to_dnf() == (~A & ~B & ~C) | ~D

    event = ~((A | ~B | C) & ~D)
    #  =>  ~(A | B | C) | ~D
    assert event.to_dnf() == (~A & B & ~C) | D

def test_dnf_factor():
    E00 = Exp(X0) > 0
    E01 = X0 < 10
    E10 = X1 < 10
    E20 = (X2**2 - X2*3) < 0
    E30 = X3 > 10
    E31 = (Sqrt(2*X3)) < 0
    E40 = X4 > 0
    E41 = X4 << [1, 5]
    E50 = 10*Log(X5) + 9 > 5

    event = (E00)
    event_dnf = event.to_dnf()
    dnf = dnf_factor(event_dnf)
    assert len(dnf) == 1
    assert dnf[0][X0] == E00

    event = E00 & E01
    event_dnf = event.to_dnf()
    dnf = dnf_factor(event_dnf)
    assert len(dnf) == 1
    assert dnf[0][X0] == E00 & E01

    event = E00 | E01
    event_dnf = event.to_dnf()
    dnf = dnf_factor(event_dnf)
    assert len(dnf) == 2
    assert dnf[0][X0] == E00
    assert dnf[1][X0] == E01

    event = E00 | (E01 & E10)
    event_dnf = event.to_dnf()
    dnf = dnf_factor(event_dnf, {X0: 0, X1: 0})
    assert len(dnf) == 2
    assert dnf[0][0] == E00
    assert dnf[1][0] == E01 & E10

    event = (E00 & E01 & E10 & E30 & E40) | (E20 & E50 & E31 & ~E41)
    # For the second clause we have:
    #   ~E41 = (-oo, 1) U (1, 5) U (5, oo)
    # so the second clause becomes
    # = (E20 & E50 & E31 & ((-oo, 1) U (1, 5) U (5, oo)))
    # = (E20 & E50 & E31 & (-oo, 1))
    #   or (E20 & E50 & E31 & (1, 5))
    #   or (E20 & E50 & E31 & (5, oo))
    event_dnf = event.to_dnf()
    event_factor = dnf_factor(event_dnf)
    assert len(event_factor) == 4
    # clause 0
    assert len(event_factor[0]) == 4
    assert event_factor[0][X0] == E00 & E01
    assert event_factor[0][X1] == E10
    assert event_factor[0][X3] == E30
    assert event_factor[0][X4] == E40
    # clause 1
    assert len(event_factor[1]) == 4
    assert event_factor[1][X3] == E31
    assert event_factor[1][X2] == E20
    assert event_factor[1][X4] == (X4 < 1)
    assert event_factor[1][X5] == E50
    # clause 2
    assert len(event_factor[2]) == 4
    assert event_factor[2][X3] == E31
    assert event_factor[2][X2] == E20
    assert event_factor[2][X4] == (1 < (X4 < 5))
    assert event_factor[2][X5] == E50
    # clause 3
    assert len(event_factor[3]) == 4
    assert event_factor[3][X3] == E31
    assert event_factor[3][X2] == E20
    assert event_factor[3][X4] == (5 < X4)
    assert event_factor[3][X5] == E50

def test_dnf_factor_1():
    A = Exp(X0) > 0
    B = X0 < 10
    C = X1 < 10
    D = X2 < 0

    event = A & B & C & ~D
    event_dnf = event.to_dnf()
    event_factor = dnf_factor(event_dnf, {X0:0, X1:0, X2:0, X3:1, X4:1, X5:2})
    assert len(event_factor) == 1
    assert event_factor[0][0] == event

def test_dnf_factor_2():
    A = X0 < 1
    B = X4 < 1
    C = X5 < 1
    event = A & B & C
    event_dnf = event.to_dnf()
    event_factor = dnf_factor(event_dnf, {X0:0, X1:0, X2:0, X3:1, X4:1, X5:2})
    assert len(event_factor) == 1
    assert event_factor[0][0] == A
    assert event_factor[0][1] == B
    assert event_factor[0][2] == C

def test_dnf_factor_3():
    A = (Exp(X0) > 0)
    B = X0 < 10
    C = X1 < 10
    D = X4 > 0
    E = (X2**2 - 3*X2) << (0, 10, 100)
    F = (10*Log(X5) + 9) > 5
    G = X4 < 4

    event = (A & B & C & ~D) | (E & F & G)
    event_dnf = event.to_dnf()
    event_factor = dnf_factor(event_dnf, {X0:0, X1:0, X2:0, X3:1, X4:1, X5:2})
    assert len(event_factor) == 2
    assert event_factor[0][0] == A & B & C
    assert event_factor[0][1] == ~D
    assert event_factor[1][0] == E
    assert event_factor[1][1] == G
    assert event_factor[1][2] == F

def test_dnf_non_disjoint_clauses():
    X = Id('X')
    Y = Id('Y')
    Z = Id('Z')

    event = (X > 0) | (Y < 0)
    overlaps = dnf_non_disjoint_clauses(event)
    assert overlaps == {1: [0]}

    event = (X > 0) | ((X < 0) & (Y < 0))
    overlaps = dnf_non_disjoint_clauses(event)
    assert not overlaps

    event = ((X > 0) & (Z < 0)) | ((X < 0) & (Y < 0)) | ((X > 1))
    overlaps = dnf_non_disjoint_clauses(event)
    assert overlaps == {2: [0]}

    event = ((X > 0) & (Z < 0)) | ((X < 0) & (Y < 0)) | ((X > 1) & (Z > 1))
    overlaps = dnf_non_disjoint_clauses(event)
    assert not overlaps

    event = ((X**2 < 9)) | (1 < X)
    overlaps = dnf_non_disjoint_clauses(event)
    assert overlaps == {1: [0]}

    event = ((X**2 < 9) & (0 < X < 1)) | (1 < X)
    overlaps = dnf_non_disjoint_clauses(event)
    assert not overlaps

def test_event_to_disjiont_union_numerical():
    X = Id('X')
    Y = Id('Y')
    Z = Id('Z')

    for event  in [
        (X > 0) | (X < 3),
        (X > 0) | (Y < 3),
        ((X > 0) & (Y < 1)) | ((X < 1) & (Y < 3)) | (Z < 0),
        ((X > 0) & (Y < 1)) | ((X < 1) & (Y < 3)) | (Z < 0) | ~(X <<{1, 3}),
    ]:
        event_dnf = dnf_to_disjoint_union(event)
        assert not dnf_non_disjoint_clauses(event_dnf)

def test_event_to_disjoint_union_nominal():
    X = Id('X')
    Y = Id('Y')
    event = (X << {'1'}) | (X << {'1', '2'})
    assert dnf_to_disjoint_union(event) == X << {'1', '2'}

    event = (X << {'1'}) | ~(Y << {'1'})
    assert dnf_to_disjoint_union(event) == EventOr([
        (X << {'1'}), ~(Y << {'1'}) & ~(X << {'1'})
    ])

def test_event_to_disjoint_union_five():
    # This test case can be visualized as follows:
    # fig, ax = plt.subplots()
    # ax.add_patch(Rectangle((2, 5), 6, 3, fill=False))
    # ax.add_patch(Rectangle((3, 7), 1, 4, fill=False))
    # ax.add_patch(Rectangle((3.5, 2), 2, 7, fill=False))
    # ax.add_patch(Rectangle((4.5, 1), 2, 5, fill=False))
    # ax.add_patch(Rectangle((5, 7), 2, 3, fill=False))
    #
    # // ((2 < X < 8) & (5 < Y < 8))
    # // ((3 < X < 4) & (8 <= Y < 11))
    # // ((3.5 < X < 5.5) & (2 < Y <= 5))
    # // ((5.5 <= X < 6.5) & (1 < Y <= 5))
    # // ((4.5 < X < 5.5) & (1 < Y <= 2))
    # // ((5 < X < 7) & (8 <= Y < 10))
    #
    # fig, ax = plt.subplots()
    # ax.add_patch(Rectangle((2, 5), 6, 3, fill=False))
    # ax.add_patch(Rectangle((3, 8), 1, 3, fill=False))
    # ax.add_patch(Rectangle((3.5, 2), 2, 3, fill=False))
    # ax.add_patch(Rectangle((5.5, 1), 1, 4, fill=False))
    # ax.add_patch(Rectangle((4.5, 1), 1, 1, fill=False))
    # ax.add_patch(Rectangle((5, 8), 2, 2, fill=False))
    X = Id('X')
    Y = Id('Y')
    E1 = ((2 < X) < 8) & ((5 < Y) < 8)
    E2 = ((3 < X) < 4) & ((7 < Y) < 11)
    E3 = ((3.5 < X) < 5.5) & ((2 < Y) < 7)
    E4 = ((4.5 < X) < 6.5) & ((1 < Y) < 6)
    E5 = ((5 < X) < 7) & ((7 < Y) < 10)
    event = E1 | E2 | E3 | E4 | E5
    dnf_to_disjoint_union(event)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from collections import ChainMap
from collections import Counter
from collections import OrderedDict
from fractions import Fraction
from functools import reduce
from inspect import getfullargspec
from itertools import chain
from math import exp
from math import log

from .dnf import dnf_factor
from .dnf import dnf_normalize
from .dnf import dnf_to_disjoint_union

from .math_util import allclose
from .math_util import flip
from .math_util import float_to_int
from .math_util import int_or_isinf_neg
from .math_util import int_or_isinf_pos
from .math_util import isinf_neg
from .math_util import logdiffexp
from .math_util import logflip
from .math_util import lognorm
from .math_util import logsumexp
from .math_util import random

from .sym_util import are_disjoint
from .sym_util import are_identical
from .sym_util import get_union
from .sym_util import partition_finite_real_contiguous
from .sym_util import partition_list_blocks
from .sym_util import sympify_number

from .transforms import EventFiniteNominal
from .transforms import EventFiniteReal
from .transforms import EventOr
from .transforms import Id

from .sets import EmptySet
from .sets import FiniteNominal
from .sets import FiniteReal
from .sets import Interval
from .sets import Range
from .sets import Union

inf = float('inf')

def memoize(f):
    table = f.__name__.split('_')[0]
    def f_(*args):
        (spn, event_factor, memo) = args
        if memo is False:
            return f(spn, event_factor_to_event, memo)
        m = getattr(memo, table)
        key = spn.get_memo_key(event_factor)
        if key not in m:
            m[key] = f(spn, event_factor, memo)
        return m[key]
    return f_

# ==============================================================================
# SPN (base class).

class SPN():
    env = None             # Environment mapping symbols to transforms.
    def __init__(self):
        raise NotImplementedError()
    def sample(self, N, prng=None):
        raise NotImplementedError()
    def sample_subset(self, symbols, N, prng=None):
        raise NotImplementedError()
    def sample_func(self, func, N, prng=None):
        raise NotImplementedError()
    def transform(self, symbol, expr):
        raise NotImplementedError()
    def logprob(self, event, memo=None):
        raise NotImplementedError()
    def prob(self, event):
        lp = self.logprob(event)
        return exp(lp)
    def condition(self, event, memo=None):
        raise NotImplementedError()
    def logpdf(self, assignment, memo=None):
        raise NotImplementedError()
    def pdf(self, assignment):
        lp = self.logpdf(assignment)
        return exp(lp)
    def mutual_information(self, A, B, memo=None):
        if memo is None:
            memo = Memo()
        lpA1 = self.logprob(A)
        lpB1 = self.logprob(B)
        lpA0 = logdiffexp(0, lpA1)
        lpB0 = logdiffexp(0, lpB1)
        lp11 = self.logprob(A & B, memo)
        lp10 = self.logprob(A & ~B, memo)
        lp01 = self.logprob(~A & B, memo)
        # lp00 = self.logprob(~A & ~B, memo)
        lp00 = logdiffexp(0, logsumexp([lp11, lp10, lp01]))
        m11 = exp(lp11) * (lp11 - (lpA1 + lpB1)) if not isinf_neg(lp11) else 0
        m10 = exp(lp10) * (lp10 - (lpA1 + lpB0)) if not isinf_neg(lp10) else 0
        m01 = exp(lp01) * (lp01 - (lpA0 + lpB1)) if not isinf_neg(lp01) else 0
        m00 = exp(lp00) * (lp00 - (lpA0 + lpB0)) if not isinf_neg(lp00) else 0
        return m11 + m10 + m01 + m00

    def __rmul__number(self, x):
        x_val = sympify_number(x)
        if not 0 < x < 1:
            raise ValueError('Weight %s must be in (0, 1)' % (str(x),))
        return PartialSumSPN([self], [x_val])
    def __rmul__(self, x):
        # Try to multiply x as a number.
        try:
            return self.__rmul__number(x)
        except TypeError:
            pass
        # Failed.
        return NotImplemented
    def __mul__(self, x):
        return x * self

    def __and__spn(self, x):
        if isinstance(x, PartialSumSPN):
            raise TypeError()
        if not isinstance(x, SPN):
            raise TypeError()
        return ProductSPN([self, x])
    def __and__(self, x):
        # Try to & x as a SPN.
        try:
            return self.__and__spn(x)
        except TypeError:
            pass
        # Failed.
        return NotImplemented

    def get_memo_key(self, event_factor):
        x = id(self)
        y = tuple(event_factor.items()) \
                if isinstance(event_factor, dict) \
                else tuple(tuple(d.items()) for d in event_factor)
        return (x, y)

# ==============================================================================
# Branch SPN.

class BranchSPN(SPN):
    symbols = None
    children = None
    def get_symbols(self):
        return self.symbols
    def logprob(self, event, memo=None):
        if memo is None:
            memo = Memo()
        event_dnf = dnf_normalize(event)
        if event_dnf is None:
            return -inf
        event_factor = dnf_factor(event_dnf)
        return self.logprob_factored(event_factor, memo)
    def condition(self, event, memo=None):
        if memo is None:
            memo = Memo()
        event_dnf = dnf_normalize(event)
        if event_dnf is None:
            raise ValueError('Zero probability event: %s' % (event,))
        if isinstance(event_dnf, EventOr):
            conjunctions = [dnf_factor(e) for e in event_dnf.subexprs]
            logps = [self.logprob_factored(c, memo) for c in conjunctions]
            indexes = [i for i, lp in enumerate(logps) if not isinf_neg(lp)]
            if not indexes:
                raise ValueError('Zero probability event: %s' % (event,))
            event_dnf = EventOr([event_dnf.subexprs[i] for i in indexes])
        event_disjoint = dnf_to_disjoint_union(event_dnf)
        event_factor = dnf_factor(event_disjoint)
        return self.condition_factored(event_factor, memo)
    def logpdf(self, assignment, memo=None):
        if memo is None:
            memo = Memo()
        return self.logpdf_factored(assignment, memo)
    def logprob_factored(self, event_factor, memo):
        raise NotImplementedError()
    def condition_factored(self, event_factor, memo):
        raise NotImplementedError()
    def logpdf_factored(self, assignment, memo):
        raise NotImplementedError()

# ==============================================================================
# Sum SPN.

class SumSPN(BranchSPN):
    """Weighted mixture of SPNs."""

    def __init__(self, children, weights):
        assert len(children) == len(weights)
        self.children = tuple(chain.from_iterable([
            spn.children
                if isinstance(spn, type(self)) else [spn]
            for spn in children
        ]))
        self.weights = tuple(chain.from_iterable([
            [weight + w for w in spn.weights]
                if isinstance(spn, type(self)) else [weight]
            for spn, weight in zip(children, weights)
        ]))
        # Derived attributes.
        self.indexes = tuple(range(len(self.weights)))
        assert allclose(float(logsumexp(weights)),  0)

        symbols = [spn.get_symbols() for spn in self.children]
        if not are_identical(symbols):
            syms = '\n'.join([', '.join(sorted(str(x) for x in s)) for s in symbols])
            raise ValueError('Mixture must have identical symbols:\n%s' % (syms,))
        self.symbols = self.children[0].get_symbols()

    def sample(self, N, prng=None):
        f_sample = lambda i, n: self.children[i].sample(n, prng=prng)
        return self.sample_many(f_sample, N, prng=prng)

    def sample_subset(self, symbols, N, prng=None):
        f_sample = lambda i, n : \
            self.children[i].sample_subset(symbols, n, prng=prng)
        return self.sample_many(f_sample, N, prng=prng)

    def sample_func(self, func, N, prng=None):
        f_sample = lambda i, n : self.children[i].sample_func(func, n, prng=prng)
        return self.sample_many(f_sample, N, prng=prng)

    def sample_many(self, func, N, prng=None):
        selections = logflip(self.weights, self.indexes, N, prng)
        counts = Counter(selections)
        samples = [func(i, counts[i]) for i in counts]
        random(prng).shuffle(samples)
        return list(chain.from_iterable(samples))

    def transform(self, symbol, expr):
        children = [spn.transform(symbol, expr) for spn in self.children]
        return SumSPN(children, self.weights)

    @memoize
    def logprob_factored(self, event_factor, memo):
        logps = [spn.logprob_factored(event_factor, memo) for spn in self.children]
        logp = logsumexp([p + w for (p, w) in zip(logps, self.weights)])
        return logp

    @memoize
    def condition_factored(self, event_factor, memo):
        logps_condt = [spn.logprob_factored(event_factor, memo) for spn in self.children]
        indexes = [i for i, lp in enumerate(logps_condt) if not isinf_neg(lp)]
        if not indexes:
            raise ValueError('Conditioning event "%s" has probability zero' % (str(event_factor),))
        logps_joint = [logps_condt[i] + self.weights[i] for i in indexes]
        children = [self.children[i].condition_factored(event_factor, memo) for i in indexes]
        weights = lognorm(logps_joint)
        return SumSPN(children, weights) if len(indexes) > 1 else children[0]

    @memoize
    def logpdf_factored(self, assignment, memo):
        # FIXME: Check base measures of children.
        logps = [spn.logpdf_factored(assignment, memo) for spn in self.children]
        return logsumexp([p + w for (p, w) in zip(logps, self.weights)])

    def __eq__(self, x):
        return isinstance(x, type(self)) \
            and self.children == x.children \
            and self.weights == x.weights
    def __hash__(self):
        x = (self.__class__, self.children, self.weights)
        return hash(x)

class ExposedSumSPN(SumSPN):
    def __init__(self, children, spn_weights):
        """Weighted mixture of SPNs with exposed internal choice."""
        assert isinstance(spn_weights, NominalLeaf)
        weights = [
            spn_weights.logprob(spn_weights.symbol << {n})
            for n in spn_weights.support
        ]
        children = [
            ProductSPN([
                NominalLeaf(spn_weights.symbol, {str(n): 1}),
                children[n]
            ]) for n in spn_weights.support
        ]
        super().__init__(children, weights)

class PartialSumSPN(SPN):
    """Weighted mixture of SPNs that do not yet sum to unity."""
    def __init__(self, children, weights):
        self.children = children
        self.weights = weights
        self.indexes = list(range(len(self.weights)))
        assert sum(weights) <  1

        symbols = [spn.get_symbols() for spn in children]
        if not are_identical(symbols):
            raise ValueError('Mixture must have identical symbols.')
        self.symbols = self.children[0].get_symbols()

    def __and__(self, x):
        raise TypeError('Weights do not sum to one.')
    def __rand__(self, x):
        raise TypeError('Weights do not sum to one.')
    def __mul__(self, x):
        raise TypeError('Cannot multiply PartialSumSPN by constant.')
    def __rmul__(self, x):
        raise TypeError('Cannot multiply PartialSumSPN by constant.')

    def __or__partialsum(self, x):
        if not isinstance(x, PartialSumSPN):
            raise TypeError()
        weights = self.weights + x.weights
        cumsum = float(sum(weights))
        if allclose(cumsum, 1):
            weights = [log(w) for w in weights]
            children = self.children + x.children
            return SumSPN(children, weights)
        if cumsum < 1:
            children = self.children + x.children
            return PartialSumSPN(children, weights)
        raise ValueError('Weights sum to more than one.')
    def __or__(self, x):
        # Try to | x as a PartialSumSPN
        try:
            return self.__or__partialsum(x)
        except TypeError:
            pass
        # Failed.
        return NotImplemented

def spn_simplify_sum(spn):
    if isinstance(spn.children[0], LeafSPN):
        return spn_simplify_sum_leaf(spn)
    if isinstance(spn.children[0], ProductSPN):
        return spn_simplify_sum_product(spn)
    assert False, 'Invalid children of SumSPN: %s' % (spn.children,)

def spn_simplify_sum_leaf(spn):
    assert all(isinstance(c, LeafSPN) for c in spn.children)
    partition = partition_list_blocks(spn.children)
    if len(partition) == len(spn.children):
        return spn
    if len(partition) == 1:
        return spn.children[0]
    children = [spn.children[block[0]] for block in partition]
    weights = [logsumexp([spn.weights[i] for i in block]) for block in partition]
    return SumSPN(children, weights)

def spn_simplify_sum_product(spn):
    # TODO: Handle case when some children are leaves with environments;
    # e.g. SumSPN([X & Y, (X & Y=X**2)])
    if not all(isinstance(c, ProductSPN) for c in spn.children):
        return spn
    children_list = [c.children for c in spn.children]
    children_simplified, weight_simplified = reduce(
        lambda state, cw: spn_simplify_sum_product_helper(state, cw[0], cw[1]),
        zip(children_list[1:], spn.weights[1:]),
        (children_list[0], spn.weights[0]),
    )
    assert allclose(logsumexp(weight_simplified), 0)
    return spn_list_to_product(children_simplified)

def spn_simplify_sum_product_helper(state, children_b, w_b):
    (children_a, w_a) = state
    weights_sum = lognorm([w_a, w_b])
    weight_overall = logsumexp([w_a, w_b])
    overlap = [(i, j)
        for j, cb in enumerate(children_b)
        for i, ca in enumerate(children_a)
        if ca == cb
    ]
    if len(overlap) == 0:
        product_a = spn_list_to_product(children_a)
        product_b = spn_list_to_product(children_b)
        children_simplified = [SumSPN([product_a, product_b], weights_sum)]
    elif len(overlap) == len(children_a):
        children_simplified = children_a
    else:
        dup_b = set(p[1] for p in overlap)
        dup_a = set(p[0] for p in overlap)
        uniq_children_b = [c for j, c in enumerate(children_b) if j not in dup_b]
        uniq_children_a = [c for i, c in enumerate(children_a) if i not in dup_a]
        dup_children = [c for i, c in enumerate(children_a) if i in dup_a]
        product_a = spn_list_to_product(uniq_children_a)
        product_b = spn_list_to_product(uniq_children_b)
        sum_a_b = SumSPN([product_a, product_b], weights_sum)
        children_simplified = [sum_a_b] + dup_children
    return (children_simplified, weight_overall)

# ==============================================================================
# Product base class.

class ProductSPN(BranchSPN):
    """List of independent SPNs."""

    def __init__(self, children):
        self.children = tuple(chain.from_iterable([
            (spn.children if isinstance(spn, type(self)) else [spn])
            for spn in children
        ]))
        # Derived attributes.
        symbols = [spn.get_symbols() for spn in self.children]
        if not are_disjoint(symbols):
            syms = '\n'.join([', '.join(sorted(str(x) for x in s)) for s in symbols])
            raise ValueError('Product must have disjoint symbols:\n%s' % (syms,))
        self.lookup = {s:i for i, syms in enumerate(symbols) for s in syms}
        self.symbols = frozenset(get_union(symbols))

    def sample(self, N, prng=None):
        samples = [spn.sample(N, prng=prng) for spn in self.children]
        return merge_samples(samples)

    def sample_subset(self, symbols, N, prng=None):
        # Partition symbols by lookup.
        index_to_symbols = {}
        for symbol in symbols:
            key = self.lookup[symbol]
            if key not in index_to_symbols:
                index_to_symbols[key] = []
            index_to_symbols[key].append(symbol)
        # Obtain the samples.
        samples = [
            self.children[i].sample_subset(symbols_i, N, prng=prng)
            for i, symbols_i in index_to_symbols.items()
        ]
        # Merge the samples.
        return merge_samples(samples)

    def sample_func(self, func, N, prng=None):
        symbols = func_symbols(self, func)
        samples = self.sample_subset(symbols, N, prng=prng)
        return func_evaluate(self, func, samples)

    def transform(self, symbol, expr):
        # TODO: This algorithm does not handle the case that expr has symbols
        # belonging to different children.  The correct solution is to
        # implement an environment in the Product, and perform substitution
        # on the event (recursively, unfortunately).
        expr_symbols = expr.get_symbols()
        assert all(e in self.get_symbols() for e in expr_symbols)
        index = [
            i for i, spn in enumerate(self.children)
            if all(s in spn.get_symbols() for s in expr_symbols)
        ]
        assert len(index) == 1, 'No child has all symbols in: %s' % (expr,)
        children = list(self.children)
        children[index[0]] = children[index[0]].transform(symbol, expr)
        return ProductSPN(children)

    @memoize
    def logprob_factored(self, event_factor, memo):
        # Adopting Inclusion--Exclusion principle for DNF event:
        # https://cp-algorithms.com/combinatorics/inclusion-exclusion.html#toc-tgt-4
        (logps_pos, logps_neg) = ([], [])
        indexes = range(len(event_factor))
        stack = [([i], i) for i in indexes]
        avoid = []
        while stack:
            # Obtain the next subset.
            subset, index = stack.pop(0)
            # Skip descendants of this subset if it contains a bad subset.
            if any(
                    len(b) <= len(subset) and all(z in subset for z in b)
                    for b in avoid):
                continue
            # Compute the probability of this subset.
            logprob = self.logprob_conjunction(event_factor, subset, memo)
            (logps_pos if len(subset) % 2 else logps_neg).append(logprob)
            # Skip descendants of this subset if measure zero.
            if isinf_neg(logprob):
                avoid.append(subset)
            # Add all subsets for which this subset is a prefix.
            for i in range(index + 1, len(indexes)):
                stack.append((subset + [indexes[i]], i))
        # Aggregate positive term.
        logp_pos = logsumexp(logps_pos)
        if isinf_neg(logp_pos) or not logps_neg:
            return logp_pos
        # Aggregate negative terms and return the difference.
        logp_neg = logsumexp(logps_neg) if logps_neg else -inf
        return logdiffexp(logp_pos, logp_neg)

    @memoize
    def condition_factored(self, event_factor, memo):
        logps = [self.logprob_conjunction([c], [0], memo) for c in event_factor]
        assert allclose(logsumexp(logps), self.logprob_factored(event_factor, memo))
        indexes = [i for (i, lp) in enumerate(logps) if not isinf_neg(lp)]
        if not indexes:
            raise ValueError('Conditioning event "%s" has probability zero'
                % (str(event_factor),))
        weights = lognorm([logps[i] for i in indexes])
        childrens = [self.condition_clause(event_factor[i], memo) for i in indexes]
        products = [ProductSPN(children) for children in childrens]
        if len(indexes) == 1:
            spn = products[0]
        else:
            spn_sum = SumSPN(products, weights)
            spn = spn_simplify_sum(spn_sum)
        return spn

    def logprob_conjunction(self, event_factor, J, memo):
        # Return probability of conjunction of |J| conjunctions.
        keys = set(self.lookup[s] for j in J for s in event_factor[j])
        return sum(
            self.logprob_conjunction_key(event_factor, J, key, memo)
            for key in keys
        )

    def logprob_conjunction_key(self, event_factor, J, key, memo):
        # Return probability of conjunction of |J| conjunction, for given key.
        clause = {}
        for j in J:
            for symbol, event in event_factor[j].items():
                if self.lookup[symbol] == key:
                    if symbol not in clause:
                        clause[symbol] = event
                    else:
                        clause[symbol] &= event
        if not clause:
            return -inf
        return self.children[key].logprob_factored((clause,), memo)

    def condition_clause(self, clause, memo):
        # Return children conditioned on a clause (one conjunction).
        children = []
        for spn in self.children:
            spn_condition = spn
            symbols = spn.get_symbols().intersection(clause)
            if symbols:
                spn_clause = ({symbol: clause[symbol] for symbol in symbols},)
                spn_condition = spn.condition_factored(spn_clause, memo)
            children.append(spn_condition)
        return children

    @memoize
    def logpdf_factored(self, assignment, memo):
        assignments = {}
        for symbol, value in assignment.items():
            key = self.lookup[symbol]
            if key not in assignments:
                assignments[key] = dict()
            assignments[key][symbol] = value
        return sum(self.children[k].logpdf_factored(a, memo)
            for k, a in assignments.items())

    def __eq__(self, x):
        return isinstance(x, type(self)) \
            and self.children == x.children
    def __hash__(self):
        x = (self.__class__, self.children)
        return hash(x)

def spn_list_to_product(children):
    return children[0] if len(children) == 1 else ProductSPN(children)

# ==============================================================================
# Basic Distribution base class.

class LeafSPN(SPN):
    symbol = None          # Symbol (Id) of base random variable
    def get_symbols(self):
        return frozenset(self.env)
    def sample(self, N, prng=None):
        return self.sample_subset(self.get_symbols(), N, prng=prng)
    def sample_subset(self, symbols, N, prng=None):
        assert all(s in self.get_symbols() for s in symbols)
        samples = self.sample__(N, prng)
        if symbols == {self.symbol}:
            return samples
        simulations = [{}] * N
        for i, sample in enumerate(samples):
            simulations[i] = dict()
            # Topological order guaranteed by OrderedDict.
            for symbol in self.env:
                sample[symbol] = self.env[symbol].evaluate(sample)
                if symbol in symbols:
                    simulations[i][symbol] = sample[symbol]
        return simulations
    def sample_func(self, func, N, prng=None):
        samples = self.sample(N, prng=prng)
        return func_evaluate(self, func, samples)
    def logprob(self, event, memo=None):
        event_subs = event.substitute(self.env)
        assert all(s in self.env for s in event.get_symbols())
        assert event_subs.get_symbols() == {self.symbol}
        if memo is None or memo is False:
            return self.logprob__(event_subs)
        key = self.get_memo_key(({self.symbol: event_subs},))
        if key not in memo.logprob:
            memo.logprob[key] = self.logprob__(event_subs)
        return memo.logprob[key]
    def condition(self, event, memo=None):
        event_subs = event.substitute(self.env)
        assert all(s in self.env for s in event.get_symbols())
        assert event_subs.get_symbols() == {self.symbol}
        if memo is None or memo is False:
            return self.condition__(event_subs)
        key = self.get_memo_key(({self.symbol: event_subs},))
        if key not in memo.condition:
            memo.condition[key] = self.condition__(event_subs)
        return memo.condition[key]
    def logprob_factored(self, event_factor, memo):
        if memo is False:
            event = event_factor_to_event(event_factor)
            return self.logprob(event)
        key = self.get_memo_key(event_factor)
        if key not in memo.logprob:
            event = event_factor_to_event(event_factor)
            memo.logprob[key] = self.logprob(event)
        return memo.logprob[key]
    def condition_factored(self, event_factor, memo):
        if memo is False:
            event = event_factor_to_event(event_factor)
            return self.condition(event)
        key = self.get_memo_key(event_factor)
        if key not in memo.condition:
            event = event_factor_to_event(event_factor)
            memo.condition[key] = self.condition(event)
        return memo.condition[key]
    def logpdf(self, assignment, memo=None):
        if memo is None:
            memo = Memo()
        return self.logpdf_factored(assignment, memo)
    @memoize
    def logpdf_factored(self, assignment, memo):
        assert len(assignment) == 1
        [(k, v)] = assignment.items()
        assert k == self.symbol
        return self.logpdf__(v)
    def sample__(self, N, prng):
        raise NotImplementedError()
    def logprob__(self, event):
        raise NotImplementedError()
    def condition__(self, event):
        raise NotImplementedError()
    def logpdf__(self, x):
        raise NotImplementedError()

# ==============================================================================
# RealLeaf base class.

class RealLeaf(LeafSPN):
    """Base class for distribution with a cumulative distribution function."""

    def __init__(self, symbol, dist, support, conditioned=None, env=None):
        assert isinstance(symbol, Id)
        assert isinstance(support, Interval)
        self.symbol = symbol
        self.dist = dist
        self.support = support
        self.conditioned = conditioned
        self.env = env or OrderedDict([(symbol, symbol)])
        # Attributes to be populated by child classes.
        self.xl = None
        self.xu = None
        self.Fl = None
        self.Fu = None
        self.logFl = None
        self.logFu = None
        self.logZ = None

    def transform(self, symbol, expr):
        assert symbol not in self.env
        assert all(s in self.env for s in expr.get_symbols())
        env = OrderedDict(self.env)
        env[symbol] = expr
        return (type(self))(self.symbol, self.dist, self.support,
            self.conditioned, env)

    def sample__(self, N, prng):
        if self.conditioned:
            # XXX Method not guaranteed to be numerically stable, see e.g,.
            # https://www.iro.umontreal.ca/~lecuyer/myftp/papers/truncated-normal-book-chapter.pdf
            # Also consider using CDF for left tail and SF for right tail.
            # Example: X ~ N(0,1) can sample X | (X < -10) but not X | (X > 10).
            u = random(prng).uniform(size=N)
            u_interval = u*self.Fl + (1-u) * self.Fu
            xs = self.dist.ppf(u_interval)
        else:
            # Simulation by vanilla inversion sampling.
            xs = self.dist.rvs(size=N, random_state=prng)
        # Wrap result in a dictionary.
        return [{self.symbol : x} for x in xs]

    def logcdf(self, x):
        if not self.conditioned:
            return self.dist.logcdf(x)
        if self.xu < x:
            return 0
        elif x < self.xl:
            return -inf
        p = logdiffexp(self.dist.logcdf(x), self.logFl)
        return p - self.logZ

    def logprob__(self, event):
        interval = event.solve()
        values = self.support & interval
        return self.logprob_values__(values)

    def logprob_values__(self, values):
        if values is EmptySet:
            return -inf
        if isinstance(values, FiniteReal):
            return self.logprob_finite__(values)
        if isinstance(values, Interval):
            return self.logprob_interval__(values)
        if isinstance(values, Union):
            logps = [self.logprob_values__(v) for v in values.args]
            return logsumexp(logps)
        assert False, 'Unknown set type: %s' % (values,)

    def logprob_finite__(self, values):
        raise NotImplementedError()
    def logprob_interval__(self, values):
        raise NotImplementedError()

    def flatten_values_contiguous(self, values):
        if isinstance(values, Interval):
            return [values]
        if isinstance(values, FiniteReal):
            assert isinstance(self, DiscreteLeaf)
            blocks = partition_finite_real_contiguous(values)
            return [Range(min(v), max(v)) for v in blocks]
        if isinstance(values, Union):
            subvalues = (self.flatten_values_contiguous(v) for v in values)
            return list(chain(*subvalues))
        assert False

    def condition__(self, event):
        interval = event.solve()
        values_set = self.support & interval
        weight = self.logprob_values__(values_set)
        # Probability zero event.
        if isinf_neg(weight):
            raise ValueError('Conditioning event "%s" has probability zero'
                % (str(event)))
        # Condition on support.
        if values_set == self.support:
            return self
        # Flatten the set.
        values = self.flatten_values_contiguous(values_set)
        # Condition on a single contiguous set.
        if len(values) == 0:
            return (type(self))(self.symbol, self.dist, values[0], True, self.env)
        # Condition on a union of contiguous set.
        else:
            weights_unorm = [self.logprob_values__(v) for v in values]
            indexes = [i for i, w in enumerate(weights_unorm) if not isinf_neg(w)]
            if not indexes:
                raise ValueError('Conditioning event "%s" has probability zero'
                    % (str(event),))
            # TODO: Normalize the weights with greater precision, e.g.,
            # https://stats.stackexchange.com/questions/66616/converting-normalizing-very-small-likelihood-values-to-probability
            weights = lognorm([weights_unorm[i] for i in indexes])
            children = [
                (type(self))(self.symbol, self.dist, values[i], True, self.env)
                for i in indexes
            ]
            return SumSPN(children, weights) if 1 < len(indexes) else children[0]
        # Unknown set.
        assert False, 'Unknown set type: %s' % (values,)

    def __hash__(self):
        d = (self.dist.dist.name, self.dist.args, tuple(self.dist.kwds.items()))
        e = tuple(self.env.items())
        x = (self.__class__, self.symbol, d, self.support, self.conditioned, e)
        return hash(x)
    def __eq__(self, x):
        return isinstance(x, type(self)) \
            and self.symbol == x.symbol \
            and self.dist.dist.name == x.dist.dist.name \
            and self.dist.args == x.dist.args \
            and self.dist.kwds == x.dist.kwds \
            and self.support == x.support \
            and self.conditioned == x.conditioned \
            and self.env == x.env

# ==============================================================================
# Continuous RealLeaf.

class ContinuousLeaf(RealLeaf):
    """Non-atomic distribution with a cumulative distribution function."""
    def __init__(self, symbol, dist, support, conditioned=None, env=None):
        super().__init__(symbol, dist, support, conditioned, env)
        self.xl = float(support.left)
        self.xu = float(support.right)
        if conditioned:
            self.Fl = self.dist.cdf(self.xl)
            self.Fu = self.dist.cdf(self.xu)
            self.logFl = self.dist.logcdf(self.xl)
            self.logFu = self.dist.logcdf(self.xu)
            self.logZ = logdiffexp(self.logFu, self.logFl)
        else:
            self.logFl = -inf
            self.logFu = 0
            self.Fl = 0
            self.Fu = 1
            self.logZ = 1

    def logpdf__(self, x):
        if isinstance(x, str):
            return -float('inf')
        xf = float(x)
        if not self.conditioned:
            return self.dist.logpdf(xf)
        if x not in self.support:
            return -inf
        return self.dist.logpdf(xf) - self.logZ

    def logprob_finite__(self, values):
        return -inf

    def logprob_interval__(self, values):
        xl = float(values.left)
        xu = float(values.right)
        logFl = self.logcdf(xl)
        logFu = self.logcdf(xu)
        return logdiffexp(logFu, logFl)

# ==============================================================================
# Discrete RealLeaf.

class DiscreteLeaf(RealLeaf):
    """Integral atomic distribution with a cumulative distribution function."""

    def __init__(self, symbol, dist, support, conditioned=None, env=None):
        super().__init__(symbol, dist, support, conditioned, env)
        assert int_or_isinf_neg(support.left)
        assert int_or_isinf_pos(support.right)
        self.xl = float_to_int(support.left) + 1*bool(support.left_open)
        self.xu = float_to_int(support.right) - 1*bool(support.right_open)
        if conditioned:
            self.Fl = self.dist.cdf(self.xl - 1)
            self.Fu = self.dist.cdf(self.xu)
            self.logFl = self.dist.logcdf(self.xl - 1)
            self.logFu = self.dist.logcdf(self.xu)
            self.logZ = logdiffexp(self.logFu, self.logFl)
        else:
            self.logFl = -inf
            self.logFu = 0
            self.Fl = 0
            self.Fu = 1
            self.logZ = 1

    def logpdf__(self, x):
        if isinstance(x, str):
            return -float('inf')
        xf = float(x)
        if not self.conditioned:
            return self.dist.logpmf(xf)
        if (x < self.xl) or (self.xu < x):
            return -inf
        return self.dist.logpmf(xf) - self.logZ

    def logprob_finite__(self, values):
        logps = [self.logpdf__(x) for x in values]
        return logps[0] if len(logps) == 1 else logsumexp(logps)
    def logprob_interval__(self, values):
        offsetl = not values.left_open and int_or_isinf_neg(values.left)
        offsetr = values.right_open and int_or_isinf_pos(values.right)
        xl = float_to_int(values.left) - offsetl
        xu = float_to_int(values.right) - offsetr
        logFl = self.logcdf(xl)
        logFu = self.logcdf(xu)
        return logdiffexp(logFu, logFl)

# ==============================================================================
# Nominal distribution.

class NominalLeaf(LeafSPN):
    """Atomic distribution, no cumulative distribution function."""

    def __init__(self, symbol, dist):
        assert isinstance(symbol, Id)
        assert all(isinstance(x, str) for x in dist)
        self.symbol = symbol
        self.dist = {x: Fraction(w) for x, w in dist.items()}
        # Derived attributes.
        self.env = {symbol: symbol}
        self.support = FiniteNominal(*dist.keys())
        self.outcomes = list(self.dist.keys())
        self.weights = list(self.dist.values())
        assert allclose(float(sum(self.weights)),  1)

    def logpdf__(self, x):
        if x not in self.dist:
            return -inf
        w = self.dist[x]
        return log(w.numerator) - log(w.denominator)

    def transform(self, symbol, expr):
        raise ValueError('Cannot transform Nominal: %s %s' % (symbol, expr))

    def sample__(self, N, prng):
        # TODO: Replace with FLDR.
        xs = flip(self.weights, self.outcomes, N, prng)
        return [{self.symbol: x} for x in xs]

    def logprob__(self, event):
        solution = event.solve()
        values = self.support & solution
        if values is EmptySet:
            return -inf
        if values == FiniteNominal(b=True):
            return 0
        p_event = sum(self.dist[x] for x in values)
        return log(p_event) if p_event != 0 else -inf

    def condition__(self, event):
        solution = event.solve()
        values = self.support & solution
        if values is EmptySet:
            raise ValueError('Zero probability condition %s' % (event,))
        p_event = sum([self.dist[x] for x in values])
        if p_event == 0:
            raise ValueError('Zero probability condition %s' % (event,))
        if p_event == 1:
            return self
        dist = {
            str(x) : (self.dist[x] / p_event) if x in values else 0
            for x in self.support
        }
        return NominalLeaf(self.symbol, dist)

    def __hash__(self):
        x = (self.__class__, self.symbol, tuple(self.dist.items()))
        return hash(x)
    def __eq__(self, x):
        return isinstance(x, type(self)) \
            and self.symbol == x.symbol \
            and self.dist == x.dist

# ==============================================================================
# Utilities.

class Memo():
    def __init__(self):
        self.logprob = {}
        self.condition = {}
        self.logpdf = {}

def spn_cache_duplicate_subtrees(spn, memo):
    if isinstance(spn, LeafSPN):
        if spn not in memo:
            memo[spn] = spn
        return memo[spn]
    if isinstance(spn, BranchSPN):
        if spn not in memo:
            memo[spn] = spn
            spn.children = list(spn.children)
            for i, c in enumerate(spn.children):
                spn.children[i] = spn_cache_duplicate_subtrees(c, memo)
            spn.children = tuple(spn.children)
        return memo[spn]
    assert False, '%s is not an spn' % (spn,)

def func_evaluate(spn, func, samples):
    args = func_symbols(spn, func)
    sample_kwargs = [{X.token: s[X] for X in args} for s in samples]
    return [func(**kwargs) for kwargs in sample_kwargs]

def func_symbols(spn, func):
    symbols = spn.get_symbols()
    args = [Id(a) for a in getfullargspec(func).args]
    unknown = [a for a in args if a not in symbols]
    if unknown:
        raise ValueError('Unknown function arguments "%s" (allowed %s)'
            % (unknown, symbols))
    return args

def merge_samples(samples):
    # input [[{X:1, Y:2}, {X:0, Y:1}], [{Z:0}, {Z:1}]] (N=2)
    # output [{X:1, Y:2, Z:0}, {X:0, Y:1, Z:1}]
    return [dict(ChainMap(*sample_list)) for sample_list in zip(*samples)]

def event_factor_to_event(event_factor):
    conjunctions = (
        reduce(lambda x, e: x & e, conjunction.values())
        for conjunction in event_factor
    )
    return reduce(lambda x, e: x | e, conjunctions)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import log

import pytest

from sppl.spn import ContinuousLeaf
from sppl.spn import PartialSumSPN
from sppl.spn import ProductSPN
from sppl.spn import SumSPN

from sppl.distributions import gamma
from sppl.distributions import norm
from sppl.transforms import Id

from sppl.math_util import allclose

X = Id('X')
Y = Id('Y')
Z = Id('Z')

def test_mul_leaf():
    for y in [0.3 * (X >> norm()), (X >> norm()) * 0.3]:
        assert isinstance(y, PartialSumSPN)
        assert len(y.weights) == 1
        assert allclose(float(sum(y.weights)), 0.3)

def test_sum_leaf():
    # Cannot sum leaves without weights.
    with pytest.raises(TypeError):
        (X >> norm()) | (X >> gamma(a=1))
    # Cannot sum a leaf with a partial sum.
    with pytest.raises(TypeError):
        0.3*(X >> norm()) | (X >> gamma(a=1))
    # Cannot sum a leaf with a partial sum.
    with pytest.raises(TypeError):
        (X >> norm()) | 0.3*(X >> gamma(a=1))
    # Wrong symbol.
    with pytest.raises(ValueError):
        0.4*(X >> norm()) | 0.6*(Y >> gamma(a=1))
    # Sum exceeds one.
    with pytest.raises(ValueError):
        0.4*(X >> norm()) | 0.7*(Y >> gamma(a=1))

    y = 0.4*(X >> norm()) | 0.3*(X >> gamma(a=1))
    assert isinstance(y, PartialSumSPN)
    assert len(y.weights) == 2
    assert allclose(float(y.weights[0]), 0.4)
    assert allclose(float(y.weights[1]), 0.3)

    y = 0.4*(X >> norm()) | 0.6*(X >> gamma(a=1))
    assert isinstance(y, SumSPN)
    assert len(y.weights) == 2
    assert allclose(float(y.weights[0]), log(0.4))
    assert allclose(float(y.weights[1]), log(0.6))
    # Sum exceeds one.
    with pytest.raises(TypeError):
        y | 0.7 * (X >> norm())

    y = 0.4*(X >> norm()) | 0.3*(X >> gamma(a=1)) | 0.1*(X >> norm())
    assert isinstance(y, PartialSumSPN)
    assert len(y.weights) == 3
    assert allclose(float(y.weights[0]), 0.4)
    assert allclose(float(y.weights[1]), 0.3)
    assert allclose(float(y.weights[2]), 0.1)

    y = 0.4*(X >> norm()) | 0.3*(X >> gamma(a=1)) | 0.3*(X >> norm())
    assert isinstance(y, SumSPN)
    assert len(y.weights) == 3
    assert allclose(float(y.weights[0]), log(0.4))
    assert allclose(float(y.weights[1]), log(0.3))
    assert allclose(float(y.weights[2]), log(0.3))

    with pytest.raises(TypeError):
        (0.3)*(0.3*(X >> norm()))
    with pytest.raises(TypeError):
        (0.3*(X >> norm())) * (0.3)
    with pytest.raises(TypeError):
        0.3*(0.3*(X >> norm()) | 0.5*(X >> norm()))

    w = 0.3*(0.4*(X >> norm()) | 0.6*(X >> norm()))
    assert isinstance(w, PartialSumSPN)

def test_product_leaf():
    with pytest.raises(TypeError):
        0.3*(X >> gamma(a=1)) & (X >> norm())
    with pytest.raises(TypeError):
        (X >> norm()) & 0.3*(X >> gamma(a=1))
    with pytest.raises(ValueError):
        (X >> norm()) & (X >> gamma(a=1))

    y = (X >> norm()) & (Y >> gamma(a=1)) & (Z >> norm())
    assert isinstance(y, ProductSPN)
    assert len(y.children) == 3
    assert y.get_symbols() == frozenset([X, Y, Z])

def test_sum_of_sums():
    w \
        = 0.3*(0.4*(X >> norm()) | 0.6*(X >> norm())) \
        | 0.7*(0.1*(X >> norm()) | 0.9*(X >> norm()))
    assert isinstance(w, SumSPN)
    assert len(w.children) == 4
    assert allclose(float(w.weights[0]), log(0.3) + log(0.4))
    assert allclose(float(w.weights[1]), log(0.3) + log(0.6))
    assert allclose(float(w.weights[2]), log(0.7) + log(0.1))
    assert allclose(float(w.weights[3]), log(0.7) + log(0.9))

    w \
        = 0.3*(0.4*(X >> norm()) | 0.6*(X >> norm())) \
        | 0.2*(0.1*(X >> norm()) | 0.9*(X >> norm()))
    assert isinstance(w, PartialSumSPN)
    assert allclose(float(w.weights[0]), 0.3)
    assert allclose(float(w.weights[1]), 0.2)

    a = w | 0.5*(X >> gamma(a=1))
    assert isinstance(a, SumSPN)
    assert len(a.children) == 5
    assert allclose(float(a.weights[0]), log(0.3) + log(0.4))
    assert allclose(float(a.weights[1]), log(0.3) + log(0.6))
    assert allclose(float(a.weights[2]), log(0.2) + log(0.1))
    assert allclose(float(a.weights[3]), log(0.2) + log(0.9))
    assert allclose(float(a.weights[4]), log(0.5))

    # Wrong symbol.
    with pytest.raises(ValueError):
        z = w | 0.4*(Y >> gamma(a=1))

def test_or_and():
    with pytest.raises(ValueError):
        (0.3*(X >> norm()) | 0.7*(Y >> gamma(a=1))) & (Z >> norm())
    a = (0.3*(X >> norm()) | 0.7*(X >> gamma(a=1))) & (Z >> norm())
    assert isinstance(a, ProductSPN)
    assert isinstance(a.children[0], SumSPN)
    assert isinstance(a.children[1], ContinuousLeaf)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest

from sppl.compilers.sppl_to_python import SPPL_Compiler

isclose = lambda a, b : abs(a-b) < 1e-10

# Test errors in visit_Assign:
# - [x] overwrite variable / array
# - [x] unknown sample target array
# - [x] assigning fresh variables in else
# - [x] non-array variable in for
# - [x] array target must not be subscript
# - [x] assign invalid Python after sampling
# - [x] assign invalid Python non-global
# - [x] invalid left-hand side (tuple, unpacking, etc)

overwrite_cases = [
'X ~= norm(); X ~= bernoulli(p=.5)',
'X = array(5); X ~= bernoulli(p=.5)',
'X ~= bernoulli(p=.5); X = array(5)',
'''
X ~= norm()
if (X > 0):
    X ~= norm()
else:
    X ~= norm()
''',
'X = array(5); W ~= norm(); X = array(10)',
]
@pytest.mark.parametrize('case', overwrite_cases)
def test_error_assign_overwrite_variable(case):
    with pytest.raises(AssertionError):
        SPPL_Compiler(case)

def test_error_assign_unknown_array():
    source = '''
X ~= uniform(loc=0, scale=1);
Y[0] ~= bernoulli(p=.1)
'''
    with pytest.raises(AssertionError):
        SPPL_Compiler(source)
    SPPL_Compiler('Y = array(10);\n%s' % (source,))

def test_error_assign_fresh_variable_else():
    source = '''
X ~= norm()
if (X > 0):
    Y ~= 2*X
else:
    W ~= 2*X
'''
    with pytest.raises(AssertionError):
        SPPL_Compiler(source)
    SPPL_Compiler(source.replace('W', 'Y'))

def test_error_assign_non_array_in_for():
    source = '''
Y ~= norm();
for i in range(10):
    X ~= norm()
'''
    with pytest.raises(AssertionError):
        SPPL_Compiler(source)
    source_prime = 'X = array(5);\n%s' % (source.replace('X', 'X[i]'),)
    SPPL_Compiler(source_prime)

def test_error_assign_array_subscript():
    source = '''
Y = array(5)
Y[0] = array(2)
'''
    with pytest.raises(AssertionError):
        SPPL_Compiler(source)
    SPPL_Compiler(source.replace('array(2)', 'norm()'))

def test_error_assign_py_constant_after_sampling():
    source = '''
X ~= norm()
Y = "foo"
'''
    with pytest.raises(AssertionError):
        SPPL_Compiler(source)

def test_error_assign_py_constant_non_global():
    source = '''
X ~= norm()
if (X > 0):
    Y = "ali"
else:
    Y = norm()
'''
    with pytest.raises(AssertionError):
        SPPL_Compiler(source)

def test_error_assign_assign_invalid_lhs():
    with pytest.raises(AssertionError):
        SPPL_Compiler('[Y] ~= norm()')
    with pytest.raises(AssertionError):
        SPPL_Compiler('(X, Y) ~= norm()')

# Test errors in visit_For:
# - [x] func is not range.
# - [x] func range has > 2 arguments
# - [x] more than one iteration variable

def test_error_for_func_not_range():
    source = '''
X = array(10)
for i in xrange(10):
    X[i] ~= norm()
'''
    with pytest.raises(AssertionError):
        SPPL_Compiler(source)
    SPPL_Compiler(source.replace('xrange', 'range'))

def test_error_for_range_step():
    source = '''
X = array(10)
X[0] ~= norm()
for i in range(1, 10, 2):
    X[i] ~= 2*X[i-1]
'''
    with pytest.raises(AssertionError):
        SPPL_Compiler(source)
    SPPL_Compiler(source.replace(', 2)' , ')'))

def test_error_for_range_multiple_vars():
    source = '''
X = array(10)
X[0] ~= norm()
for (i, j) in range(1, 9):
    X[i] ~= 2*X[i-1]
'''
    with pytest.raises(AssertionError):
        SPPL_Compiler(source)
    SPPL_Compiler(source.replace('(i, j)' , 'i'))

# Test errors in visit_If:
# - [x] if without matching else/elif
def test_error_if_no_else():
    source = '''
X ~= norm()
if (X > 0):
    Y ~= 2*X + 1
'''
    with pytest.raises(AssertionError):
        SPPL_Compiler(source)
    SPPL_Compiler('%s\nelse:\n    Y~= norm()' % (source,))
    SPPL_Compiler('%s\nelif (X<0):\n    Y~= norm()' % (source,))

# Test SPPL_Transformer
# Q = Z in {0, 1}         E = Z << {0, 1}
# Q = Z not in {0, 1}     Q = ~(Z << {0, 1})
# Q = Z == 'foo'          Q = Z << {'foo'}
# B = Z != 'foo'          Q = ~(Z << {'foo'})

def test_transform_in():
    source = '''
X ~= choice({'foo': .5, 'bar': .1, 'baz': .4})
Y = X in {'foo', 'baz'}
'''
    compiler = SPPL_Compiler(source)
    py_source = compiler.render_module()
    assert 'X << {\'foo\', \'baz\'}' in py_source
    assert 'X in' not in py_source

def test_transform_in_not():
    source = '''
X ~= choice({'foo': .5, 'bar': .1, 'baz': .4})
Y = X not in {'foo', 'baz'}
'''
    compiler = SPPL_Compiler(source)
    py_source = compiler.render_module()
    assert '~ (X << {\'foo\', \'baz\'})' in py_source
    assert 'not in' not in py_source

def test_transform_eq():
    source = '''
X ~= choice({'foo': .5, 'bar': .1, 'baz': .4})
Y = X == 'foo'
'''
    compiler = SPPL_Compiler(source)
    py_source = compiler.render_module()
    assert 'X << {\'foo\'}' in py_source
    assert '==' not in py_source

def test_transform_eq_not():
    source = '''
X ~= choice({'foo': .5, 'bar': .1, 'baz': .4})
Y = X != 'foo'
'''
    compiler = SPPL_Compiler(source)
    py_source = compiler.render_module()
    assert '~ (X << {\'foo\'})' in py_source
    assert '!=' not in py_source

def test_compile_all_constructs():
    source = '''
X = array(10)
W = array(10)
Y = randint(low=1, high=2)
Z = bernoulli(p=0.1)

E = choice({'1': 0.3, '2': 0.7})


for i in range(1,5):
    W[i] = uniform(loc=0, scale=2)
    X[i] = bernoulli(p=0.5)

X[0] ~= gamma(a=1)
H = (X[0]**2 + 2*X[0] + 3)**(1, 10)

# Here is a comment, with indentation on next line
    


X[5] ~= 0.3*atomic(loc=0) | 0.4*atomic(loc=-1) | 0.3*atomic(loc=3)
if X[5] == 0:
    X[7] = bernoulli(p=0.1)
    X[8] = 1 + X[3]
elif (X[5]**2 == 1):
    X[7] = bernoulli(p=0.2)
    X[8] = 1 + X[3]
else:
    if (X[3] in {0, 1}):
        X[7] = bernoulli(p=0.2)
        X[8] = 1 + X[3]
    else:
        X[7] = bernoulli(p=0.2)
        X[8] = 1 + X[3]
'''
    compiler = SPPL_Compiler(source)
    namespace = compiler.execute_module()
    model = namespace.model
    assert isclose(model.prob(namespace.X[5] << {0}), .3)
    assert isclose(model.prob(namespace.X[5] << {-1}), .4)
    assert isclose(model.prob(namespace.X[5] << {3}), .3)
    assert isclose(model.prob(namespace.E << {'1'}), .3)
    assert isclose(model.prob(namespace.E << {'2'}), .7)

def test_imports():
    source = '''
Y ~= bernoulli(p=.5)
Z ~= choice({str(i): Fraction(1, 5) for i in range(5)})
X = array(5)
for i in range(5):
    X[i] ~= Fraction(1,2) * Y
'''
    compiler = SPPL_Compiler(source)
    with pytest.raises(NameError):
        compiler.execute_module()
    compiler = SPPL_Compiler('from fractions import Fraction\n%s' % (source,))
    namespace = compiler.execute_module()
    for i in range(5):
        assert isclose(namespace.model.prob(namespace.Z << {str(i)}), .2)

def test_ifexp():
    source = '''
from fractions import Fraction
Y ~= choice({str(i): Fraction(1, 4) for i in range(4)})
Z ~= (
    atomic(loc=0)    if (Y in {'0', '1'}) else
    atomic(loc=4)    if (Y == '2') else
    atomic(loc=6))
'''
    compiler = SPPL_Compiler(source)
    assert 'IfElse' in compiler.render_module()
    namespace = compiler.execute_module()
    assert isclose(namespace.model.prob(namespace.Z << {0}), .5)
    assert isclose(namespace.model.prob(namespace.Z << {4}), .25)
    assert isclose(namespace.model.prob(namespace.Z << {6}), .25)

def test_switch_shallow():
    source = '''
Y ~= choice({'0': .25, '1': .5, '2': .25})

switch (Y) cases (i in ['0', '1', '2']):
    Z ~= atomic(loc=int(i))
'''
    compiler = SPPL_Compiler(source)
    namespace = compiler.execute_module()
    assert isclose(namespace.model.prob(namespace.Z << {0}), .25)
    assert isclose(namespace.model.prob(namespace.Z << {1}), .5)
    assert isclose(namespace.model.prob(namespace.Z << {2}), .25)

def test_switch_enumerate():
    source = '''
Y ~= choice({'0': .25, '1': .5, '2': .25})

switch (Y) cases (i,j in enumerate(['0', '1', '2'])):
    Z ~= atomic(loc=i+int(j))
'''
    compiler = SPPL_Compiler(source)
    namespace = compiler.execute_module()
    assert isclose(namespace.model.prob(namespace.Z << {0}), .25)
    assert isclose(namespace.model.prob(namespace.Z << {2}), .5)
    assert isclose(namespace.model.prob(namespace.Z << {4}), .25)

def test_switch_nested():
    source = '''
Y ~= randint(low=0, high=4)
W ~= randint(low=0, high=2)

switch (Y) cases (i in range(0, 5)):
    Z ~= choice({str(i): 1})
    switch (W) cases (i in range(0, 2)): V ~= atomic(loc=i)
'''
    compiler = SPPL_Compiler(source)
    namespace = compiler.execute_module()
    assert isclose(namespace.model.prob(namespace.Z << {'0'}), .25)
    assert isclose(namespace.model.prob(namespace.Z << {'1'}), .25)
    assert isclose(namespace.model.prob(namespace.Z << {'2'}), .25)
    assert isclose(namespace.model.prob(namespace.Z << {'3'}), .25)
    assert isclose(namespace.model.prob(namespace.V << {0}), .5)
    assert isclose(namespace.model.prob(namespace.V << {1}), .5)

def test_condition_simple():
    source = '''
Y ~= norm(loc=0, scale=2)
condition((0 < Y) < 2)
Z ~= binom(n=10, p=.2)
condition(Z == 0)
'''
    compiler = SPPL_Compiler(source)
    namespace = compiler.execute_module()
    assert isclose(namespace.model.prob((0 < namespace.Y) < 2), 1)

def test_condition_if():
    source = '''
Y ~= norm(loc=0, scale=2)
if (Y > 1):
    condition(Y < 1)
else:
    condition(Y > -1)
'''
    with pytest.raises(Exception):
        compiler = SPPL_Compiler(source)
        compiler.execute_module()
    compiler = SPPL_Compiler(source.replace('Y > 1', 'Y > 0'))
    namespace = compiler.execute_module()
    assert isclose(namespace.model.prob((-1 < namespace.Y) < 1), 1)

def test_constant_parameter():
    source = '''
parameters = [1, 2, 3]
n_array = 2
Y = array(n_array + 1)
for i in range(n_array + 1):
    Y[i] = randint(low=parameters[i], high=parameters[i]+1)
'''
    compiler = SPPL_Compiler(source)
    namespace = compiler.execute_module()
    assert isclose(namespace.model.prob(namespace.Y[0] << {1}), 1)
    assert isclose(namespace.model.prob(namespace.Y[1] << {2}), 1)
    assert isclose(namespace.model.prob(namespace.Y[2] << {3}), 1)
    with pytest.raises(AssertionError):
        SPPL_Compiler('%sZ = "foo"\n' % (source,))

def test_error_array_length():
    with pytest.raises(TypeError):
        SPPL_Compiler('Y = array(1.3)').execute_module()
    with pytest.raises(TypeError):
        SPPL_Compiler('Y = array(\'foo\')').execute_module()
    # Length zero array
    namespace = SPPL_Compiler('Y = array(-1)').execute_module()
    assert len(namespace.Y) == 0



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest
import sympy

from sympy import Rational as Rat

from sppl.math_util import allclose

from sppl.sets import EmptySet
from sppl.sets import FiniteNominal
from sppl.sets import FiniteReal
from sppl.sets import Interval
from sppl.sets import Reals
from sppl.sets import Union
from sppl.sets import inf as oo

from sppl.sym_util import sympy_solver
from sppl.transforms import Exp
from sppl.transforms import Identity
from sppl.transforms import Log
from sppl.transforms import Logarithm
from sppl.transforms import Sqrt

X = sympy.symbols('X')
Y = Identity('Y')

def test_solver_1_open():
    # log(x) > 2
    solution = Interval.open(sympy.exp(2), oo)
    event = Log(Y) > 2
    answer = event.solve()
    assert answer == solution

def test_solver_1_closed():
    # log(x) >= 2
    solution = Interval(sympy.exp(2), oo)
    event = Log(Y) >= 2
    answer = event.solve()
    assert answer == solution

def test_solver_2_open():
    # log(x) < 2 & (x < exp(2))
    solution = EmptySet
    event = (Log(Y) > 2) & (Y < sympy.exp(2))
    answer = event.solve()
    assert answer == solution

def test_solver_2_closed():
    # (log(x) <= 2) & (x >= exp(2))
    solution = FiniteReal(sympy.exp(2))
    event = (Log(Y) >= 2) & (Y <= sympy.exp(2))
    answer = event.solve()
    assert answer == solution

def test_solver_4():
    # (x >= 0) & (x <= 0)
    solution = Reals
    event = (Y >= 0) | (Y <= 0)
    answer = event.solve()
    assert answer == solution

def test_solver_5_open():
    # (2*x+10 < 4) & (x + 10 > 3)
    solution = Interval.open(3-10, (4-10)/2)
    event = ((2*Y + 10) < 4) & (Y + 10 > 3)
    answer = event.solve()
    assert answer == solution

def test_solver_5_ropen():
    # (2*x+10 < 4) & (x + 10 >= 3)
    solution = Interval.Ropen(3-10, (4-10)/2)
    event = ((2*Y + 10) < 4) & (Y + 10 >= 3)
    answer = event.solve()
    assert answer == solution

def test_solver_5_lopen():
    # (2*x + 10 < 4) & (x + 10 >= 3)
    solution =Interval.Lopen(3-10, (4-10)/2)
    event = ((2*Y + 10) <= 4) & (Y + 10 > 3)
    answer = event.solve()
    assert answer == solution

def test_solver_6():
    # (x**2 - 2*x) > 10
    solution =  Union(
        Interval.open(-oo, 1 - sympy.sqrt(11)),
        Interval.open(1 + sympy.sqrt(11), oo))
    event = (Y**2 - 2*Y) > 10
    answer = event.solve()
    assert answer == solution

def test_solver_7():
    # Illegal expression, cannot express in our custom DSL.
    expr = (X**2 - 2*X + sympy.exp(X)) > 10
    with pytest.raises(ValueError):
        sympy_solver(expr)

def test_solver_8():
    Z = sympy.symbols('Z')
    expr = (X + Z < 3)
    with pytest.raises(ValueError):
        sympy_solver(expr)

def test_solver_9_open():
    # 2(log(x))**3 - log(x) -5 > 0
    solution = Interval.open(
        sympy.exp(1/(6*(sympy.sqrt(2019)/36 + Rat(5,4))**(Rat(1, 3)))
            + (sympy.sqrt(2019)/36 + Rat(5,4))**(Rat(1,3))),
        oo)
    # Our solver handles this case as follows
    # expr' = 2*Z**3 - Z - 5 > 0 [[subst. Z=log(X)]]
    # [Z_low, Z_high] = sympy_solver(expr')
    #       Z_low < Z iff Z_low < log(X) iff exp(Z_low) < X
    #       Z < Z_high iff log(X) < Z_high iff X < exp(Z_high)
    # sympy_solver(expr) = [exp(Z_low), exp(Z_high)]
    # For F invertible, can thus solve Poly(coeffs, F) > 0 using this method.
    event = 2*(Log(Y))**3 - Log(Y) - 5 > 0
    answer = event.solve()
    assert answer == solution

def test_solver_9_closed():
    # 2(log(x))**3 - log(x) -5 >= 0
    solution = Interval(
        sympy.exp(1/(6*(sympy.sqrt(2019)/36 + Rat(5,4))**(Rat(1, 3)))
            + (sympy.sqrt(2019)/36 + Rat(5,4))**(Rat(1,3))),
        oo)
    event = 2*(Log(Y))**3 - Log(Y) - 5 >= 0
    answer = event.solve()
    assert answer == solution

def test_solver_10():
    # Sympy hangs on this test.
    # exp(sqrt(log(x))) > -5
    solution = Interval(1, oo)
    event = Exp(Sqrt(Log(Y))) > -5
    answer = event.solve()
    assert answer == solution

def test_solver_11_open():
    # exp(sqrt(log(x))) > 6
    solution = Interval.open(sympy.exp(sympy.log(6)**2), oo)
    event = Exp(Sqrt(Log(Y))) > 6
    answer = event.solve()
    assert answer == solution

def test_solver_11_closed():
    # exp(sqrt(log(x))) >= 6
    solution = Interval(sympy.exp(sympy.log(6)**2), oo)
    event = Exp(Sqrt(Log(Y))) >= 6
    answer = event.solve()
    assert answer == solution

def test_solver_12():
    # 2*sqrt(|x|) - 3 > 10
    solution = Union(
        Interval.open(-oo, -Rat(169, 4)),
        Interval.open(Rat(169, 4), oo))
    event = (2*Sqrt(abs(Y)) - 3) > 10
    answer = event.solve()
    assert answer == solution

def test_solver_13():
    # 2*sqrt(|x|**2) - 3 > 10
    solution = Union(
        Interval.open(-oo, -Rat(13, 2)),
        Interval.open(Rat(13, 2), oo))
    event = (2*Sqrt(abs(Y)**2) - 3) > 10
    answer = event.solve()
    assert answer == solution

def test_solver_14():
    # x**2 > 10
    solution = Union(
        Interval.open(-oo, -sympy.sqrt(10)),
        Interval.open(sympy.sqrt(10), oo))
    event = Y**2 > 10
    answer = event.solve()
    assert answer == solution

def test_solver_15():
    # ((x**4)**(1/7)) < 9
    solution = Interval.open(-27*sympy.sqrt(3), 27*sympy.sqrt(3))
    event = ((Y**4))**(Rat(1, 7)) < 9
    answer = event.solve()
    assert answer == solution

def test_solver_16():
    # (x**(1/7))**4 < 9
    solution = Interval.Ropen(0, 27*sympy.sqrt(3))
    event = ((Y**Rat(1,7)))**4 < 9
    answer = event.solve()
    assert answer == solution

@pytest.mark.xfail(reason='too slow', strict=True)
@pytest.mark.timeout(3)
def test_solver_17():
    p = sympy.Poly(
        (X - sympy.sqrt(2)/10) * (X+Rat(10, 7)) * (X - sympy.sqrt(5)),
        X)
    expr = p.args[0] < 1
    sympy_solver(expr)

def test_solver_18():
    # 3*(x**(1/7))**4 - 3*(x**(1/7))**2 <= 9
    solution = Interval(0, (Rat(1, 2) + sympy.sqrt(13)/2)**(Rat(7, 2)))

    Z = Y**(Rat(1, 7))
    expr = 3*Z**4 - 3*Z**2
    event = (expr <= 9)
    answer = event.solve()
    assert answer == solution

    interval = (~event).solve()
    assert interval == Interval.open(solution.right, oo)

def test_solver_19():
    # 3*(x**(1/7))**4 - 3*(x**(1/7))**2 <= 9
    #   or || 3*(x**(1/7))**4 - 3*(x**(1/7))**2 > 11
    solution = Union(
        Interval(0, (Rat(1, 2) + sympy.sqrt(13)/2)**(Rat(7, 2))),
        Interval.open((Rat(1,2) + sympy.sqrt(141)/6)**(Rat(7, 2)), oo))

    Z = Y**(Rat(1, 7))
    expr = 3*Z**4 - 3*Z**2
    event = (expr <= 9) | (expr > 11)
    answer = event.solve()
    assert answer == solution

    interval = (~event).solve()
    assert interval == Interval.Lopen(
        solution.args[0].right,
        solution.args[1].left)

def test_solver_20():
    # log(x**2 - 3) < 5
    solution = Union(
        Interval.open(-sympy.sqrt(3 + sympy.exp(5)), -sympy.sqrt(3)),
        Interval.open(sympy.sqrt(3), sympy.sqrt(3 + sympy.exp(5))))
    event = Log(Y**2 - 3) < 5
    answer = event.solve()
    assert answer == solution

def test_solver_21__ci_():
    # 1 <= log(x**3 - 3*x + 3) < 5
    # Can only be solved by numerical approximation of roots.
    # https://www.wolframalpha.com/input/?i=1+%3C%3D+log%28x**3+-+3x+%2B+3%29+%3C+5
    solution = Union(
        Interval(
            -1.777221448430427630375448631016427343692,
            0.09418455242255462832154474245589911789464),
        Interval.Ropen(
            1.683036896007873002053903888560528225797,
            5.448658707897512189124586716091172798465))

    expr = Log(Y**3 - 3*Y + 3)
    event = ((1 <= expr) & (expr < 5))
    answer = event.solve()
    assert isinstance(answer, Union)
    assert len(answer.args) == 2
    first = answer.args[0] if answer.args[0].a < 0 else answer.args[1]
    second = answer.args[0] if answer.args[0].a > 0 else answer.args[1]
    # Check first interval.
    assert not first.left_open
    assert not first.right_open
    assert allclose(float(first.a), float(solution.args[0].a))
    assert allclose(float(first.b), float(solution.args[0].b))
    # Check second interval.
    assert not second.left_open
    assert second.right_open
    assert allclose(float(second.a), float(solution.args[1].a))
    assert allclose(float(second.b), float(solution.args[1].b))

def test_solver_22():
    # 2 < abs(X) < 5
    event = (2 < abs(Y)) < 5
    solution = Interval.open(2, 5) | Interval.open(-5, -2)
    assert event.solve() == solution
    # 2 <= abs(X) < 5
    event = (2 <= abs(Y)) < 5
    solution = Interval.Ropen(2, 5) | Interval.Lopen(-5, -2)
    assert event.solve() == solution
    # 2 < abs(X) <= 5
    event = (2 <  abs(Y)) <= 5
    solution = Interval.Lopen(2, 5) | Interval.Ropen(-5, -2)
    assert event.solve() == solution
    # 2 <= abs(X) <= 5
    event = (2 <=  abs(Y)) <= 5
    solution = Interval(2, 5) | Interval(-5, -2)
    assert event.solve() == solution

    # -2 < abs(X) < 5
    event = (-2 < abs(Y)) < 5
    solution = Interval.open(-5, 5)
    assert event.solve() == solution
    # # -2 <= abs(X) < 5
    event = (-2 <= abs(Y)) < 5
    solution = Interval.open(-5, 5)
    assert event.solve() == solution
    # -2 < abs(X) <= 5
    event = (-2 <  abs(Y)) <= 5
    solution = Interval(-5, 5)
    assert event.solve() == solution
    # 2 <= abs(X) <= 5
    event = (-2 <=  abs(Y)) <= 5
    solution = Interval(-5, 5)
    assert event.solve() == solution

def test_solver_23_reciprocal_lte():
    for c in [1, 3]:
        # Positive
        # 1 / X < 10
        solution = Interval.Ropen(-oo, 0) | Interval.Lopen(Rat(c, 10), oo)
        event = (c / Y) < 10
        assert event.solve() == solution
        # 1 / X <= 10
        solution = Interval.Ropen(-oo, 0) | Interval(Rat(c, 10), oo)
        event = (c / Y) <= 10
        assert event.solve() == solution
        # 1 / X <= sqrt(2)
        solution = Interval.Ropen(-oo, 0) | Interval(c / sympy.sqrt(2), oo)
        event = (c / Y) <= sympy.sqrt(2)
        assert event.solve() == solution
        # Negative.
        # 1 / X < -10
        solution = Interval.open(-Rat(c, 10), 0)
        event = (c / Y) < -10
        assert event.solve() == solution
        # 1 / X <= -10
        solution = Interval.Ropen(-Rat(c, 10), 0)
        event = (c / Y) <= -10
        assert event.solve() == solution
        # 1 / X <= -sqrt(2)
        solution = Interval.Ropen(-c / sympy.sqrt(2), 0)
        event = (c / Y) <= -sympy.sqrt(2)
        assert event.solve() == solution

def test_solver_23_reciprocal_gte():
    for c in [1, 3]:
        # Positive
        # 10 < 1 / X
        solution = Interval.open(0, Rat(c, 10))
        event = 10 < (c / Y)
        assert event.solve() == solution
        # 10 <= 1 / X
        solution = Interval.Lopen(0, Rat(c, 10))
        event = 10 <= (c / Y)
        assert event.solve() == solution
        # Negative
        # -10 < 1 / X
        solution = Interval.Lopen(0, oo) | Interval.open(-oo, -Rat(c, 10))
        event = -10 < (c / Y)
        assert event.solve() == solution
        # -10 <= 1 / X
        solution = Interval.Lopen(0, oo) | Interval.Lopen(-oo, -Rat(c, 10))
        event =  -10 <= (c / Y)
        assert event.solve() == solution

def test_solver_23_reciprocal_range():
    solution = Interval.Ropen(-1, -Rat(1, 3))
    event = ((-3 < 1/Y) <= -1)
    assert event.solve() == solution

    solution = Interval.open(0, Rat(1, 3))
    event = ((-3 < 1/(2*Y-1)) < -1)
    assert event.solve() == solution

    solution = Interval.open(-1 / sympy.sqrt(3), 1 / sympy.sqrt(3))
    event = ((-3 < 1/(2*(abs(Y)**2)-1)) <= -1)
    assert event.solve() == solution

    solution = Union(
        Interval.open(-1 / sympy.sqrt(3), 0),
        Interval.open(0, 1 / sympy.sqrt(3)))
    event = ((-3 < 1/(2*(abs(Y)**2)-1)) < -1)
    assert event.solve() == solution

def test_solver_24_negative_power_integer():
    # Case 1.
    event = Y**(-3) < 6
    assert event.solve() == Union(
        Interval.open(-oo, 0),
        Interval.open(6**Rat(-1, 3), oo))
    # Case 2.
    event = (-1 < Y**(-3)) < 6
    assert event.solve() == Union(
        Interval.open(-oo, -1),
        Interval.open(6**Rat(-1, 3), oo))
    # Case 3.
    event = 5 <= Y**(-3)
    assert event.solve() == Interval.Lopen(0, 5**Rat(-1, 3))
    # Case 4.
    event = (5 <= Y**(-3)) < 6
    assert event.solve() == Interval.Lopen(6**Rat(-1, 3), 5**Rat(-1, 3))

def test_solver_24_negative_power_Rat():
    # Case 1.
    event = Y**Rat(-1, 3) < 6
    assert event.solve() == Interval.Lopen(Rat(1, 216), oo)
    # Case 2.
    event = (-1 < Y**Rat(-1, 3)) < 6
    assert event.solve() == Interval.Lopen(Rat(1, 216), oo)
    # Case 3.
    event = 5 <= Y**Rat(-1, 3)
    assert event.solve() == Interval.Lopen(0, Rat(1, 125))
    # Case 4.
    event = (5 <= Y**Rat(-1, 3)) < 6
    assert event.solve() == Interval.Lopen(Rat(1, 216), Rat(1, 125))

def test_solver_25_constant():
    event = (0*Y + 1) << {1}
    assert event.solve() == Reals
    event = (0*Y + 1) << {0}
    assert event.solve() is EmptySet
    event = (0.9 < (0*Y + 1)) < 1
    assert event.solve() is EmptySet
    event = (0.9 < (0*Y + 1)**2) <= 1
    assert event.solve() == Reals
    event = (0.9 < (0*Y + 2)**2) <= 1
    assert event.solve() is EmptySet
    event = (0*Y + 2)**2 << {4}
    assert event.solve() == Reals

def test_solver_26_piecewise_one_expr_basic_event():
    event = (Y**2)*(0 <= Y) < 2
    assert event.solve() == Interval.Ropen(0, sympy.sqrt(2))
    event = (0 <= Y)*(Y**2) < 2
    assert event.solve() == Interval.Ropen(0, sympy.sqrt(2))
    event = ((0 <= Y) < 5)*(Y < 1) << {1}
    assert event.solve() == Interval.Ropen(0, 1)
    event = ((0 <= Y) < 5)*(~(Y < 1)) << {1}
    assert event.solve() == Interval.Ropen(1, 5)
    event = 10*(0 <= Y) << {10}
    assert event.solve() == Interval(0, oo)
    event = 10*(0 <= Y) << {0}
    assert event.solve() == Interval.Ropen(-oo, 0)

def test_solver_26_piecewise_one_expr_compound_event():
    event = (Y**2)*((Y < 0) | (0 < Y)) < 2
    assert event.solve() == Union(
        Interval.open(-sympy.sqrt(2), 0),
        Interval.open(0, sympy.sqrt(2)))

def test_solver_27_piecewise_many():
    expr = (Y < 0)*(Y**2) + (0 <= Y)*Y**(Rat(1, 2))
    event = expr << {3}
    assert sorted(event.solve()) == [-sympy.sqrt(3), 9]
    event = 0 < expr
    assert event.solve() == Union(
        Interval.open(-oo, 0),
        Interval.open(0, oo))

    # TODO: Consider banning the restriction of a function
    # to a segment outside of its domain.
    expr = (Y < 0)*Y**(Rat(1, 2))
    assert (expr < 1).solve() is EmptySet

def test_solver_finite_injective():
    sqrt3 = sympy.sqrt(3)
    # Identity.
    solution = FiniteReal(2, 4, -10, sqrt3)
    event = Y << {2, 4, -10, sqrt3}
    assert event.solve() == solution
    # Exp.
    solution = FiniteReal(sympy.log(10), sympy.log(3), sympy.log(sqrt3))
    event = Exp(Y) << {10, 3, sqrt3}
    assert event.solve() == solution
    # Exp2.
    solution = FiniteReal(sympy.log(10, 2), 4, sympy.log(sqrt3, 2))
    event = (2**Y) << {10, 16, sqrt3}
    assert event.solve() == solution
    # Log.
    solution = FiniteReal(sympy.exp(10), sympy.exp(-3), sympy.exp(sqrt3))
    event = Log(Y) << {10, -3, sqrt3}
    assert event.solve() == solution
    # Log2
    solution = FiniteReal(sympy.Pow(2, 10), sympy.Pow(2, -3), sympy.Pow(2, sqrt3))
    event = Logarithm(Y, 2) << {10, -3, sqrt3}
    assert event.solve() == solution
    # Radical.
    solution = FiniteReal(7**4, 12**4, sqrt3**4)
    event = Y**Rat(1, 4) << {7, 12, sqrt3}
    assert event.solve() == solution

def test_solver_finite_non_injective():
    sqrt2 = sympy.sqrt(2)
    # Abs.
    solution = FiniteReal(-10, -3, 3, 10)
    event = abs(Y) << {10, 3}
    assert event.solve() == solution
    # Abs(Poly).
    solution = FiniteReal(-5, -Rat(3,2), Rat(3,2), 5)
    event = abs(2*Y) << {10, 3}
    assert event.solve() == solution
    # Poly order 2.
    solution = FiniteReal(-sqrt2, sqrt2)
    event = (Y**2) << {2}
    assert event.solve() == solution
    # Poly order 3.
    solution = FiniteReal(1, 3)
    event = Y**3 << {1, 27}
    assert event.solve() == solution
    # Poly Abs.
    solution = FiniteReal(-3, -1, 1, 3)
    event = (abs(Y))**3 << {1, 27}
    assert event.solve() == solution
    # Abs Not.
    solution = Union(
        Interval.open(-oo, -1),
        Interval.open(-1, 1),
        Interval.open(1, oo))
    event = ~(abs(Y) << {1})
    assert event.solve() == solution
    # Abs in EmptySet.
    solution = EmptySet
    event = (abs(Y))**3 << set()
    assert event.solve() == solution
    # Abs Not in EmptySet (yields all reals).
    solution = Interval(-oo, oo)
    event = ~(((abs(Y))**3) << set())
    assert event.solve() == solution
    # Log in Reals (yields positive reals).
    solution = Interval.open(0, oo)
    event = ~((Log(Y))**3 << set())
    assert event.solve() == solution

def test_solver_finite_symbolic():
    # Transform can never be symbolic.
    event = Y << {'a', 'b'}
    assert event.solve() == FiniteNominal('a', 'b')
    # Complement the Identity.
    event = ~(Y << {'a', 'b'})
    assert event.solve() == FiniteNominal('a', 'b', b=True)
    # Transform can never be symbolic.
    event = Y**2 << {'a', 'b'}
    assert event.solve() is EmptySet
    # Complement the Identity.
    event = ~(Y**2 << {'a', 'b'})
    assert event.solve() == FiniteNominal(b=True)
    # Solve Identity mixed.
    event = Y << {9, 'a', '7'}
    assert event.solve() == Union(
        FiniteReal(9),
        FiniteNominal('a', '7'))
    # Solve Transform mixed.
    event = Y**2 << {9, 'a', 'b'}
    assert event.solve() == FiniteReal(-3, 3)
    # Solve a disjunction.
    event = (Y << {'a', 'b'}) | (Y << {'c'})
    assert event.solve() == FiniteNominal('a', 'b', 'c')
    # Solve a conjunction with intersection.
    event = (Y << {'a', 'b'}) & (Y << {'b', 'c'})
    assert event.solve() == FiniteNominal('b')
    # Solve a conjunction with no intersection.
    event = (Y << {'a', 'b'}) & (Y << {'c'})
    assert event.solve() is EmptySet
    # Solve a disjunction with complement.
    event = (Y << {'a', 'b'}) & ~(Y << {'c'})
    assert event.solve() == FiniteNominal('a', 'b')
    # Solve a disjunction with complement.
    event = (Y << {'a', 'b'}) | ~(Y << {'c'})
    assert event.solve() == FiniteNominal('c', b=True)
    # Union of interval and symbolic.
    event = (Y**2 <= 9) | (Y << {'a'})
    assert event.solve() == Interval(-3, 3) | FiniteNominal('a')
    # Union of interval and not symbolic.
    event = (Y**2 <= 9) | ~(Y << {'a'})
    assert event.solve() == Interval(-3, 3) | FiniteNominal('a', b=True)
    # Intersection of interval and symbolic.
    event = (Y**2 <= 9) & (Y << {'a'})
    assert event.solve() is EmptySet
    # Intersection of interval and not symbolic.
    event = (Y**2 <= 9) & ~(Y << {'a'})
    assert event.solve() == EmptySet



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import log

import pytest

from sppl.distributions import DistributionMix
from sppl.distributions import bernoulli
from sppl.distributions import choice
from sppl.distributions import discrete
from sppl.distributions import norm
from sppl.distributions import poisson
from sppl.distributions import rv_discrete
from sppl.distributions import uniformd
from sppl.math_util import allclose
from sppl.sets import FiniteNominal
from sppl.sets import Interval
from sppl.sets import inf as oo
from sppl.spn import ContinuousLeaf
from sppl.spn import DiscreteLeaf
from sppl.spn import NominalLeaf
from sppl.spn import SumSPN
from sppl.transforms import Id

X = Id('X')

def test_simple_parse_real():
    assert isinstance(.3*bernoulli(p=.1), DistributionMix)
    a = .3*bernoulli(p=.1) | .5 * norm() | .2*poisson(mu=7)
    spn = a(X)
    assert isinstance(spn, SumSPN)
    assert allclose(spn.weights, [log(.3), log(.5), log(.2)])
    assert isinstance(spn.children[0], DiscreteLeaf)
    assert isinstance(spn.children[1], ContinuousLeaf)
    assert isinstance(spn.children[2], DiscreteLeaf)
    assert spn.children[0].support == Interval(0, 1)
    assert spn.children[1].support == Interval(-oo, oo)
    assert spn.children[2].support == Interval(0, oo)

def test_simple_parse_nominal():
    assert isinstance(.7 * choice({'a': .1, 'b': .9}), DistributionMix)
    a = .3*bernoulli(p=.1) | .7*choice({'a': .1, 'b': .9})
    spn = a(X)
    assert isinstance(spn, SumSPN)
    assert allclose(spn.weights, [log(.3), log(.7)])
    assert isinstance(spn.children[0], DiscreteLeaf)
    assert isinstance(spn.children[1], NominalLeaf)
    assert spn.children[0].support == Interval(0, 1)
    assert spn.children[1].support == FiniteNominal('a', 'b')

def test_error():
    with pytest.raises(TypeError):
        'a'*bernoulli(p=.1)
    a = .1  *bernoulli(p=.1) | .7*poisson(mu=8)
    with pytest.raises(Exception):
        a(X)

def test_parse_rv_discrete():
    for dist in [
        rv_discrete(values=((1, 2, 10), (.3, .5, .2))),
        discrete({1: .3, 2: .5, 10: .2})
    ]:
        spn = dist(X)
        assert spn.support == Interval(1, 10)
        assert allclose(spn.prob(X<<{1}), .3)
        assert allclose(spn.prob(X<<{2}), .5)
        assert allclose(spn.prob(X<<{10}), .2)
        assert allclose(spn.prob(X<=10), 1)

    dist = uniformd(values=((1, 2, 10, 0)))
    spn = dist(X)
    assert spn.support == Interval(0, 10)
    assert allclose(spn.prob(X<<{1}), .25)
    assert allclose(spn.prob(X<<{2}), .25)
    assert allclose(spn.prob(X<<{10}), .25)
    assert allclose(spn.prob(X<<{0}), .25)



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest

from sppl.transforms import Exp
from sppl.transforms import Id
from sppl.transforms import Log

X = Id('X')
Y = Id('Y')
W = Id('W')
Z = Id('Z')

def test_event_basic_invertible():
    expr = X**2 + 10*X

    event = (0 < expr)
    assert event.evaluate({X: 10})
    assert not (~event).evaluate({X: 10})

    event = (0 > expr)
    assert not event.evaluate({X: 10})
    assert (~event).evaluate({X: 10})

    event = (0 < expr) < 100
    for val in [0, 10]:
        assert not event.evaluate({X: val})
        assert (~event).evaluate({X: val})

    event = (0 <= expr) <= 100
    x_eq_100 = (expr << {100}).solve()
    for val in [0, list(x_eq_100)[0]]:
        assert event.evaluate({X: val})
        assert not (~event).evaluate({X: val})

    event = expr << {11}
    assert event.evaluate({X: 1})
    assert not event.evaluate({X: 3})
    assert not (~event).evaluate({X: 1})
    assert (~event).evaluate({X :3})

    with pytest.raises(ValueError):
        event.evaluate({Y: 10})

def test_event_compound():
    expr0 = abs(X)**2 + 10*abs(X)
    expr1 = Y-1

    event = (0 < expr0) | (0 < expr1)
    assert not event.evaluate({X: 0, Y: 0})
    for i in range(1, 100):
        assert event.evaluate({X: i, Y: i})
    with pytest.raises(ValueError):
        event.evaluate({X: 0})
    with pytest.raises(ValueError):
        event.evaluate({Y: 0})

    event = (100 <= expr0) & ((expr0 < 200) | (3 < expr1))
    assert not event.evaluate({X: 0, Y: 0})

    x_eq_150 = (expr0 << {150}).solve()
    assert event.evaluate({X: list(x_eq_150)[0], Y: 0})
    assert event.evaluate({X: list(x_eq_150)[1], Y: 0})

    x_eq_500 = (expr0 << {500}).solve()
    assert not event.evaluate({X: list(x_eq_500)[1], Y: 4})
    assert event.evaluate({X: list(x_eq_500)[1], Y: 5})

def test_event_solve_multi():
    event = (Exp(abs(3*X**2)) > 1) | (Log(Y) < 0.5)
    with pytest.raises(ValueError):
        event.solve()
    event = (Exp(abs(3*X**2)) > 1) & (Log(Y) < 0.5)
    with pytest.raises(ValueError):
        event.solve()



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import os
import tempfile
import time

from math import exp

import graphviz
import networkx as nx

from sppl.spn import NominalLeaf
from sppl.spn import ProductSPN
from sppl.spn import RealLeaf
from sppl.spn import SumSPN

gensym = lambda: 'r%s' % (str(time.time()).replace('.', ''),)

def render_networkx_graph(spn):
    if isinstance(spn, NominalLeaf):
        G = nx.DiGraph()
        root = gensym()
        G.add_node(root, label='%s\n%s' % (spn.symbol.token, 'Nominal'))
        return G
    if isinstance(spn, RealLeaf):
        G = nx.DiGraph()
        root = gensym()
        kwds = '\n%s' % (tuple(spn.dist.kwds.values()),) if spn.dist.kwds else ''
        G.add_node(root, label='%s\n%s%s' % (spn.symbol.token, spn.dist.dist.name, kwds))
        if len(spn.env) > 1:
            for k, v in spn.env.items():
                if k != spn.symbol:
                    roott = gensym()
                    G.add_node(roott, label=str(v), style='filled')
                    G.add_edge(root, roott, label=' %s' % (str(k),), style='dashed')
        return G
    if isinstance(spn, SumSPN):
        G = nx.DiGraph()
        root = gensym()
        G.add_node(root, label='\N{PLUS SIGN}')
        # Add nodes and edges from children.
        G_children = [render_networkx_graph(c) for c in spn.children]
        for i, x in enumerate(G_children):
            G.add_nodes_from(x.nodes.data())
            G.add_edges_from(x.edges.data())
            subroot = list(nx.topological_sort(x))[0]
            G.add_edge(root, subroot, label='%1.3f' % (exp(spn.weights[i]),))
        return G
    if isinstance(spn, ProductSPN):
        G = nx.DiGraph()
        root = gensym()
        G.add_node(root, label='\N{MULTIPLICATION SIGN}')
        # Add nodes and edges from children.
        G_children = [render_networkx_graph(c) for c in spn.children]
        for x in G_children:
            G.add_nodes_from(x.nodes.data())
            G.add_edges_from(x.edges.data())
            subroot = list(nx.topological_sort(x))[0]
            G.add_edge(root, subroot)
        return G
    assert False, 'Unknown SPN type: %s' % (spn,)

def render_graphviz(spn, filename=None, ext=None, show=None):
    fname = filename
    if filename is None:
        f = tempfile.NamedTemporaryFile(delete=False)
        fname = f.name
    G = render_networkx_graph(spn)
    ext = ext or 'png'
    assert ext in ['png', 'pdf'], 'Extension must be .pdf or .png'
    fname_dot = '%s.dot' % (fname,)
    # nx.set_edge_attributes(G, 'serif', 'fontname')
    # nx.set_node_attributes(G, 'serif', 'fontname')
    nx.nx_agraph.write_dot(G, fname_dot)
    source = graphviz.Source.from_file(fname_dot, format=ext)
    source.render(filename=fname, view=show)
    os.unlink(fname)
    return source



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest

from sppl.compilers.ast_to_spn import Condition
from sppl.compilers.ast_to_spn import Id
from sppl.compilers.ast_to_spn import Sample
from sppl.compilers.ast_to_spn import Sequence
from sppl.distributions import beta
from sppl.distributions import choice
from sppl.distributions import randint
from sppl.math_util import allclose

Y = Id('Y')
X = Id('X')

def test_condition_nominal():
    command = Sequence(
        Sample(Y, choice({'a':.1, 'b':.1, 'c':.8})),
        Condition(Y << {'a', 'b'}))
    model = command.interpret()
    assert allclose(model.prob(Y << {'a'}), .5)
    assert allclose(model.prob(Y << {'b'}), .5)
    assert allclose(model.prob(Y << {'c'}), 0)

def test_condition_real_discrete_range():
    command = Sequence(
        Sample(Y, randint(low=0, high=4)),
        Condition(Y << {0, 1}))
    model = command.interpret()
    assert allclose(model.prob(Y << {0}), .5)
    assert allclose(model.prob(Y << {1}), .5)

@pytest.mark.xfail(strict=True, reason='https://github.com/probcomp/sum-product-dsl/issues/77')
def test_condition_real_discrete_no_range():
    command = Sequence(
        Sample(Y, randint(low=0, high=4)),
        Condition(Y << {0, 2}))
    model = command.interpret()
    assert allclose(model.prob(Y << {0}), .5)
    assert allclose(model.prob(Y << {1}), .5)

def test_condition_real_continuous():
    command = Sequence(
        Sample(Y, beta(a=1, b=1)),
        Condition(Y < .5))
    model = command.interpret()
    assert allclose(model.prob(Y < .5), 1)
    assert allclose(model.prob(Y > .5), 0)

def test_condition_prob_zero():
    with pytest.raises(Exception):
        Sequence(
            Sample(Y, {'a':.1, 'b':.1, 'c':.8}),
            Condition(Y << {'d'})).interpret()
    with pytest.raises(Exception):
        Sequence(
            Sample(Y, beta(a=1, b=1)),
            Condition(Y > 1)).interpret()



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

from math import log

import pytest

from numpy import linspace

from sppl.distributions import bernoulli
from sppl.distributions import beta
from sppl.distributions import randint
from sppl.compilers.ast_to_spn import IfElse
from sppl.compilers.ast_to_spn import Sample
from sppl.compilers.ast_to_spn import Sequence
from sppl.compilers.ast_to_spn import Switch
from sppl.compilers.ast_to_spn import Id
from sppl.math_util import allclose
from sppl.math_util import logsumexp
from sppl.sym_util import binspace

Y = Id('Y')
X = Id('X')

def test_simple_model_eq():
    command_switch = Sequence(
        Sample(Y, randint(low=0, high=4)),
        Switch(Y, range(0, 4), lambda i:
            Sample(X, bernoulli(p=1/(i+1)))))
    model_switch = command_switch.interpret()

    command_ifelse = Sequence(
        Sample(Y, randint(low=0, high=4)),
        IfElse(
            Y << {0}, Sample(X, bernoulli(p=1/(0+1))),
            Y << {1}, Sample(X, bernoulli(p=1/(1+1))),
            Y << {2}, Sample(X, bernoulli(p=1/(2+1))),
            Y << {3}, Sample(X, bernoulli(p=1/(3+1))),
        ))
    model_ifelse = command_ifelse.interpret()

    for model in [model_switch, model_ifelse]:
        symbols = model.get_symbols()
        assert symbols == {X, Y}
        assert allclose(
            model.logprob(X << {1}),
            logsumexp([-log(4) - log(i+1) for i in range(4)]))

def test_simple_model_lte():
    command_switch = Sequence(
        Sample(Y, beta(a=2, b=3)),
        Switch(Y, binspace(0, 1, 5), lambda i:
            Sample(X, bernoulli(p=i.right))))
    model_switch = command_switch.interpret()

    command_ifelse = Sequence(
        Sample(Y, beta(a=2, b=3)),
        IfElse(
            Y <= 0,     Sample(X, bernoulli(p=0)),
            Y <= 0.25,  Sample(X, bernoulli(p=.25)),
            Y <= 0.50,  Sample(X, bernoulli(p=.50)),
            Y <= 0.75,  Sample(X, bernoulli(p=.75)),
            Y <= 1,     Sample(X, bernoulli(p=1)),
        ))
    model_ifelse = command_ifelse.interpret()

    grid = [float(x) for x in linspace(0, 1, 5)]
    for model in [model_switch, model_ifelse]:
        symbols = model.get_symbols()
        assert symbols == {X, Y}
        assert allclose(
            model.logprob(X << {1}),
            logsumexp([
                model.logprob((il < Y) <= ih) + log(ih)
                for il, ih in zip(grid[:-1], grid[1:])
            ]))

def test_simple_model_enumerate():
    command_switch = Sequence(
        Sample(Y, randint(low=0, high=4)),
        Switch(Y, enumerate(range(0, 4)), lambda i,j:
            Sample(X, bernoulli(p=1/(i+j+1)))))
    model = command_switch.interpret()
    assert allclose(model.prob(Y<<{0} & (X << {1})), .25 * 1/(0+0+1))
    assert allclose(model.prob(Y<<{1} & (X << {1})), .25 * 1/(1+1+1))
    assert allclose(model.prob(Y<<{2} & (X << {1})), .25 * 1/(2+2+1))
    assert allclose(model.prob(Y<<{3} & (X << {1})), .25 * 1/(3+3+1))

def test_error_range():
    with pytest.raises(AssertionError):
        # Switch cases do not sum to one.
        command = Sequence(
            Sample(Y, randint(low=0, high=4)),
            Switch(Y, range(0, 3), lambda i:
                Sample(X, bernoulli(p=1/(i+1)))))
        command.interpret()

def test_error_linspace():
    with pytest.raises(AssertionError):
        # Switch cases do not sum to one.
        command = Sequence(
            Sample(Y, beta(a=2, b=3)),
            Switch(Y, linspace(0, .5, 5), lambda i:
                Sample(X, bernoulli(p=i))))
        command.interpret()

def test_error_binspace():
    with pytest.raises(AssertionError):
        # Switch cases do not sum to one.
        command = Sequence(
            Sample(Y, beta(a=2, b=3)),
            Switch(Y, binspace(0, .5, 5), lambda i:
                Sample(X, bernoulli(p=i.right))))
        command.interpret()



# Copyright 2020 MIT Probabilistic Computing Project.
# See LICENSE.txt

import pytest

from sppl.distributions import bernoulli
from sppl.distributions import beta
from sppl.distributions import randint
from sppl.distributions import uniform

from sppl.compilers.ast_to_spn import For
from sppl.compilers.ast_to_spn import Id
from sppl.compilers.ast_to_spn import IdArray
from sppl.compilers.ast_to_spn import IfElse
from sppl.compilers.ast_to_spn import Sample
from sppl.compilers.ast_to_spn import Sequence
from sppl.compilers.ast_to_spn import Switch
from sppl.compilers.ast_to_spn import Transform

from sppl.sym_util import binspace

simAll   = Id('simAll')
sim      = IdArray('sim', 5)
p1       = IdArray('p1', 5)
p2       = IdArray('p2', 5)
clickA   = IdArray('clickA', 5)
clickB   = IdArray('clickB', 5)

ns = 5
nd = ns - 1

def get_command_randint():
    return Sequence(
        Sample(simAll, randint(low=0, high=ns)),
        For(0, 5, lambda k:
            Switch (simAll, range(0, ns), lambda i:
                Sequence (
                    Sample (sim[k], bernoulli(p=i/nd)),
                    Sample (p1[k], randint(low=0, high=ns)),
                    IfElse(
                        sim[k] << {1},
                            Sequence(
                                Transform (p2[k], p1[k]),
                                Switch (p1[k], range(ns), lambda j:
                                    Sequence(
                                        Sample(clickA[k], bernoulli(p=i/nd)),
                                        Sample(clickB[k], bernoulli(p=i/nd))))),
                        True,
                            Sequence(
                                Sample (p2[k], randint(low=0, high=ns)),
                                Switch (p1[k], range(ns), lambda j:
                                      Sample(clickA[k], bernoulli(p=j/nd))),
                                Switch (p2[k], range(ns), lambda j:
                                      Sample (clickB[k], bernoulli(p=j/nd)))))))))

def get_command_beta():
    return Sequence(
        Sample(simAll, beta(a=2, b=3)),
        For(0, 5, lambda k:
            Switch (simAll, binspace(0, 1, ns), lambda i:
                Sequence (
                    Sample (sim[k], bernoulli(p=i.right)),
                    Sample (p1[k], uniform()),
                    IfElse(
                        sim[k] << {1},
                            Sequence(
                                Transform (p2[k], p1[k]),
                                Switch (p1[k], binspace(0, 1, ns), lambda j:
                                    Sequence(
                                        Sample(clickA[k], bernoulli(p=i.right)),
                                        Sample(clickB[k], bernoulli(p=i.right))))),
                        True,
                            Sequence(
                                Sample (p2[k], uniform()),
                                Switch (p1[k], binspace(0, 1, ns), lambda j:
                                      Sample(clickA[k], bernoulli(p=j.right))),
                                Switch (p2[k], binspace(0, 1, ns), lambda j:
                                      Sample (clickB[k], bernoulli(p=j.right)))))))))

@pytest.mark.parametrize('get_command', [get_command_randint, get_command_beta])
def test_clickgraph_crash__ci_(get_command):
    command = get_command()
    model = command.interpret()
    model_condition = model.condition(
            (clickA[0] << {1}) & (clickB[0] << {1})
        &   (clickA[1] << {1}) & (clickB[1] << {1})
        &   (clickA[2] << {1}) & (clickB[2] << {1})
        &   (clickA[3] << {0}) & (clickB[3] << {0})
        &   (clickA[4] << {0}) & (clickB[4] << {0}))
    probabilities = [model_condition.prob(simAll << {i}) for i in range(ns)]
    assert all(p <= probabilities[nd-1] for p in probabilities)
